import json
import os
from typing import List, Optional
import pandas as pd
from pydantic import BaseModel
import requests
from aryaxai.client.client import APIClient
from aryaxai.common.environment import Environment
from aryaxai.core.organization import Organization
from aryaxai.common.xai_uris import (
    AVAILABLE_BATCH_SERVERS_URI,
    AVAILABLE_CUSTOM_SERVERS_URI,
    AVAILABLE_SYNTHETIC_CUSTOM_SERVERS_URI,
    CLEAR_NOTIFICATIONS_URI,
    CREATE_ORGANIZATION_URI,
    GET_CASE_PROFILE_URI,
    GET_NOTIFICATIONS_URI,
    LOGIN_URI,
    UPLOAD_DATA_PROJECT_URI,
    USER_ORGANIZATION_URI,
)
import getpass


class XAI(BaseModel):
    """Base class to connect with AryaXAI platform"""

    env: Environment = Environment()
    api_client: APIClient = APIClient()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        debug = self.env.get_debug()
        base_url = self.env.get_base_url()

        self.api_client = APIClient(debug=debug, base_url=base_url)

    def login(self):
        """login to AryaXAI platform

        :param api_key: API key, defaults to XAI_ACCESS_TOKEN environment variable
        """
        access_token = os.environ.get("XAI_ACCESS_TOKEN", None) or getpass.getpass(
            "Enter your Arya XAI Access Token: "
        )

        if not access_token:
            raise ValueError("Either set XAI_ACCESS_TOKEN or pass the Access token")

        res = self.api_client.post(LOGIN_URI, payload={"access_token": access_token})
        self.api_client.update_headers(res["access_token"])
        self.api_client.set_access_token(access_token)

        print("Authenticated successfully.")

    def organizations(self) -> pd.DataFrame:
        """Get all organizations associated with user

        :return: Organization details dataframe
        """

        res = self.api_client.get(USER_ORGANIZATION_URI)

        if not res["success"]:
            raise Exception(res.get("details", "Failed to get organizations"))

        res["details"].insert(
            0,
            {
                "name": "Personal",
                "organization_owner": True,
                "organization_admin": True,
                "current_users": 1,
                "created_by": "you",
            },
        )

        organization_df = pd.DataFrame(
            res["details"],
            columns=[
                "name",
                "organization_owner",
                "organization_admin",
                "current_users",
                "created_by",
                "created_at",
            ],
        )

        return organization_df

    def organization(self, organization_name: str) -> Organization:
        """Select specific organization

        :param organization_name: Name of the organization to be used
        :return: Organization object
        """
        if organization_name == "personal":
            return Organization(
                api_client=self.api_client,
                **{
                    "name": "Personal",
                    "organization_owner": True,
                    "organization_admin": True,
                    "current_users": 1,
                    "created_by": "you",
                }
            )

        organizations = self.api_client.get(USER_ORGANIZATION_URI)

        if not organizations["success"]:
            raise Exception(organizations.get("details", "Failed to get organizations"))

        user_organization = [
            Organization(api_client=self.api_client, **organization)
            for organization in organizations["details"]
        ]

        organization = next(
            filter(
                lambda organization: organization.name == organization_name,
                user_organization,
            ),
            None,
        )

        if organization is None:
            raise Exception("Organization Not Found")

        return organization

    def create_organization(self, organization_name: str) -> Organization:
        """Create New Organization

        :param organization_name: Name of the new organization
        :return: Organization object
        """
        payload = {"organization_name": organization_name}
        res = self.api_client.post(CREATE_ORGANIZATION_URI, payload)

        if not res["success"]:
            raise Exception(res.get("details", "Failed to create organization"))

        return Organization(api_client=self.api_client, **res["organization_details"])

    def get_notifications(self) -> pd.DataFrame:
        """get user notifications

        :return: notification details dataFrame
        """
        res = self.api_client.get(GET_NOTIFICATIONS_URI)

        if not res["success"]:
            raise Exception("Error while getting user notifications.")

        notifications = res["details"]

        if not notifications:
            return "No notifications found."

        return pd.DataFrame(notifications).reindex(
            columns=["project_name", "message", "time"]
        )

    def clear_notifications(self) -> str:
        """clear user notifications

        :return: response
        """
        res = self.api_client.post(CLEAR_NOTIFICATIONS_URI)

        if not res["success"]:
            raise Exception("Error while clearing user notifications.")

        return res["details"]

    def available_batch_servers(self) -> dict:
        """available custom batch servers

        :return: response
        """
        res = self.api_client.get(AVAILABLE_BATCH_SERVERS_URI)
        return res["details"]

    def available_custom_servers(self) -> dict:
        """available custom servers

        :return: response
        """
        res = self.api_client.get(AVAILABLE_CUSTOM_SERVERS_URI)
        return res

    def available_synthetic_custom_servers(self) -> dict:
        """available synthetic custom servers

        :return: response
        """
        res = self.api_client.get(AVAILABLE_SYNTHETIC_CUSTOM_SERVERS_URI)
        return res["details"]
    
    def register_case(
        self,
        token: str,
        client_id: str,
        unique_identifier: Optional[str] = None,
        project_name: str = None,
        tag: Optional[str] = None,
        data: Optional[str] = None,
        processed_data: Optional[bool] = False,
        merge: Optional[bool] = False,
        image_class: Optional[str] = None,
        prompt: Optional[str] = None,
        serverless_instance_type: Optional[str] = None,
        explainability_method: Optional[str] = None,
        explain_model: Optional[bool] = False,
        session_id: Optional[str] = None,
        xai: Optional[str] = None,
        file_path: Optional[str] = None
    ):
        form_data = {
            "client_id": client_id,
            "project_name": project_name,
            "unique_identifier": unique_identifier,
            "tag": tag,
            "data": json.dumps(data) if isinstance(data, list) else data,
            "processed_data": str(processed_data).lower(),
            "merge": str(merge).lower(),
            "image_class": image_class,
            "prompt": prompt,
            "serverless_instance_type": serverless_instance_type,
            "explainability_method": explainability_method,
            "explain_model": str(explain_model).lower(),
            "session_id": str(session_id).lower(),
            "xai": xai
        }
        headers = {
            "x-api-token": token
        }
        form_data = {k: v for k, v in form_data.items() if v is not None}
        files = {}
        if file_path:
            files["in_file"] = open(file_path, "rb")
        response = requests.post(
            self.env.get_base_url() + "/" + UPLOAD_DATA_PROJECT_URI,
            data=form_data,
            files=files if files else None,
            headers=headers
        ).json()
        if files:
            files["in_file"].close()
        return response

    def case_profile(
        self,
        token: str,
        client_id: str,
        unique_identifier: Optional[str] = None,
        project_name: str = None,
        tag: str = None,
        xai: Optional[List[str]] = None,
        refresh: Optional[bool] = None
    ):
        headers = {
            "x-api-token": token
        }
        payload = {
            "client_id": client_id,
            "project_name": project_name,
            "unique_identifier": unique_identifier,
            "tag": tag,
            "xai": xai,
            "refresh": refresh
        }
        res = requests.post(
            self.env.get_base_url() + "/" + GET_CASE_PROFILE_URI,
            headers=headers,
            json=payload
        ).json()
        return res["details"]



from typing import Any, Callable, Dict, List, Optional, Tuple, TypedDict, Union 

class Guard:
    """Predefined guardrail configurations for common use cases"""
    
    @staticmethod
    def detect_pii(entities: List[str] = None) -> Dict[str, Any]:
        """Template for PII detection guardrail
        
        Args:
            entities: List of PII entity types to detect
        """
        if entities is None:
            entities = ["EMAIL_ADDRESS", "PHONE_NUMBER", "PERSON", "ADDRESS"]
        
        return {
            "name": "Detect PII",
            "config": {
                "pii_entities": entities
            }
        }
    
    @staticmethod
    def nsfw_text(threshold: float = 0.8, validation_method: str = "sentence") -> Dict[str, Any]:
        """Template for NSFW text detection guardrail
        
        Args:
            threshold: Confidence threshold for detection (0.0-1.0)
            validation_method: "sentence", "paragraph", or "document"
        """
        return {
            "name": "NSFW Text",
            "config": {
                "threshold": threshold,
                "validation_method": validation_method
            }
        }
    @staticmethod
    def ban_list(banned_words: List[str]) -> Dict[str, Any]:
        """Template for banned words guardrail"""
        return {
            "name": "Ban List",
            "config": {
                "banned_words": banned_words
            }
        }
    @staticmethod
    def bias_check(threshold: float = 0.9) -> Dict[str, Any]:
        """Template for bias check guardrail"""
        return {
            "name": "Bias Check",
            "config": {
                "threshold": threshold
            }
        }
    
    @staticmethod
    def competitor_check(competitors: List[str]) -> Dict[str, Any]:
        """Template for competitor guardrail"""
        return {
            "name": "Competitor Check",
            "config": {
                "competitors": competitors
            }
        }
    
    @staticmethod
    def correct_language(expected_language_iso: str = "en", threshold: float = 0.75) -> Dict[str, Any]:
        """Template for correct language guardrail"""
        return {
            "name": "Correct Language",
            "config": {
                "expected_language_iso": expected_language_iso,
                "threshold": threshold
            }
        }
    
    @staticmethod
    def gibberish_text(threshold: float = 0.5, validation_method: str = "sentence") -> Dict[str, Any]:
        """Template for gibberish text guardrail"""
        return {
            "name": "Gibberish Text",
            "config": {
                "threshold": threshold,
                "validation_method": validation_method
            }
        }
    
    @staticmethod
    def profanity_free() -> Dict[str, Any]:
        """Template for profanity free guardrail"""
        return {
            "name": "Profanity Free",
            "config": {}
        }
    
    @staticmethod
    def secrets_present() -> Dict[str, Any]:
        """Template for secrets present guardrail"""
        return {
            "name": "Secrets Present",
            "config": {}
        }
    
    @staticmethod
    def toxic_language(threshold: float = 0.5, validation_method: str = "sentence") -> Dict[str, Any]:
        """Template for toxic language guardrail"""
        return {
            "name": "Toxic Language",
            "config": {
                "threshold": threshold,
                "validation_method": validation_method
            }
        }

    @staticmethod
    def contains_string(substring: str) -> Dict[str, Any]:
        """Template for contains string guardrail"""
        return {
            "name": "Contains String",
            "config": {
                "substring": substring
            }
        }

    @staticmethod
    def detect_jailbreak(threshold: float = 0.0) -> Dict[str, Any]:
        """Template for detect jailbreak guardrail"""
        return {
            "name": "Detect Jailbreak",
            "config": {
                "threshold": threshold
            }
        }

    @staticmethod
    def endpoint_is_reachable() -> Dict[str, Any]:
        """Template for endpoint is reachable guardrail"""
        return {
            "name": "Endpoint Is Reachable",
            "config": {}
        }
    
    @staticmethod
    def ends_with(end: str) -> Dict[str, Any]:
        """Template for ends with guardrail"""
        return {
            "name": "Ends With",
            "config": {
                "end": end
            }
        }

    @staticmethod
    def has_url() -> Dict[str, Any]:
        """Template for has url guardrail"""
        return {
            "name": "Has Url",
            "config": {}
        }

    @staticmethod
    def lower_case() -> Dict[str, Any]:
        """Template for lower case guardrail"""
        return {
            "name": "Lower Case",
            "config": {}
        }

    @staticmethod
    def mentions_drugs() -> Dict[str, Any]:
        """Template for mentions drugs guardrail"""
        return {
            "name": "Mentions Drugs",
            "config": {}
        }

    @staticmethod
    def one_line() -> Dict[str, Any]:
        """Template for one line guardrail"""
        return {
            "name": "One Line",
            "config": {}
        }

    @staticmethod
    def reading_time(reading_time: float) -> Dict[str, Any]:
        """Template for reading time guardrail"""
        return {
            "name": "Reading Time",
            "config": {
                "reading_time": reading_time
            }
        }

    @staticmethod
    def redundant_sentences(threshold: int = 70) -> Dict[str, Any]:
        """Template for redundant sentences guardrail"""
        return {
            "name": "Redundant Sentences",
            "config": {
                "threshold": threshold
            }
        }

    @staticmethod
    def regex_match(regex: str, match_type: str = "search") -> Dict[str, Any]:
        """Template for regex match guardrail"""
        return {
            "name": "Regex Match",
            "config": {
                "regex": regex,
                "match_type": match_type
            }
        }

    @staticmethod
    def sql_column_presence(cols: List[str]) -> Dict[str, Any]:
        """Template for SQL column presence guardrail"""
        return {
            "name": "Sql Column Presence",
            "config": {
                "cols": cols
            }
        }

    @staticmethod
    def two_words() -> Dict[str, Any]:
        """Template for two words guardrail"""
        return {
            "name": "Two Words",
            "config": {}
        }

    @staticmethod
    def upper_case() -> Dict[str, Any]:
        """Template for upper case guardrail"""
        return {
            "name": "Upper Case",
            "config": {}
        }

    @staticmethod
    def valid_choices(choices: List[str]) -> Dict[str, Any]:
        """Template for valid choices guardrail"""
        return {
            "name": "Valid Choices",
            "config": {
                "choices": choices
            }
        }

    @staticmethod
    def valid_json() -> Dict[str, Any]:
        """Template for valid json guardrail"""
        return {
            "name": "Valid Json",
            "config": {}
        }

    @staticmethod
    def valid_length(min: Optional[int] = None, max: Optional[int] = None) -> Dict[str, Any]:
        """Template for valid length guardrail"""
        config = {}
        if min is not None:
            config['min'] = min
        if max is not None:
            config['max'] = max
        return {
            "name": "Valid Length",
            "config": config
        }

    @staticmethod
    def valid_range(min: Optional[int] = None, max: Optional[int] = None) -> Dict[str, Any]:
        """Template for valid range guardrail"""
        config = {}
        if min is not None:
            config['min'] = min
        if max is not None:
            config['max'] = max
        return {
            "name": "Valid Range",
            "config": config
        }

    @staticmethod
    def valid_url() -> Dict[str, Any]:
        """Template for valid url guardrail"""
        return {
            "name": "Valid URL",
            "config": {}
        }
    
    @staticmethod
    def web_sanitization() -> Dict[str, Any]:
        """Template for web sanitization guardrail"""
        return {
            "name": "Web Sanitization",
            "config": {}
        }


import inspect

from agents import (
    Agent,
    AgentOutputSchema,
    AgentOutputSchemaBase,
    AgentSpanData,
    AgentsException,
    Handoff,
    Model,
    ModelBehaviorError,
    ModelResponse,
    OutputGuardrail,
    OutputGuardrailResult,
    OutputGuardrailTripwireTriggered,
    RunConfig,
    RunContextWrapper,
    RunErrorDetails,
    RunHooks,
    RunItem,
    RunResult,
    RunResultStreaming,
    Session,
    Span,
    SpanError,
    Tool,
    Usage,
    agent_span,
    trace,
)
from openai.types.responses import ResponseCompletedEvent
from openai.types.responses.response_prompt_param import (
    ResponsePromptParam,
)
from agents.exceptions import (
    AgentsException,
    InputGuardrailTripwireTriggered,
    MaxTurnsExceeded,
    ModelBehaviorError,
    OutputGuardrailTripwireTriggered,
    RunErrorDetails,
    UserError,
)
from agents.guardrail import (
    InputGuardrail,
    InputGuardrailResult,
    OutputGuardrail,
    OutputGuardrailResult,
)

from agents.handoffs import Handoff, HandoffInputFilter, handoff

from typing import Any, cast, Generic
from agents.run import (
    TResponseInputItem,
    TContext,
    DEFAULT_MAX_TURNS,
    RunOptions,
    AgentToolUseTracker,
    _error_tracing,
    TraceCtxManager,
    get_model_tracing_impl,
    AgentToolUseTracker,
    NextStepFinalOutput,
    NextStepHandoff,
    NextStepRunAgain,
    QueueCompleteSentinel,
    RunImpl,
    SingleStepResult,
    TraceCtxManager,
)
from agents.stream_events import AgentUpdatedStreamEvent
from agents.items import ItemHelpers, ModelResponse, RunItem, TResponseInputItem
from agents.lifecycle import RunHooks
from agents.logger import logger
from agents.memory import Session
from agents.model_settings import ModelSettings
from agents.models.interface import Model, ModelProvider
from agents.models.multi_provider import MultiProvider
from agents.result import RunResult, RunResultStreaming
from agents.run_context import RunContextWrapper, TContext
from agents.stream_events import AgentUpdatedStreamEvent, RawResponsesStreamEvent
from agents.tool import Tool
from agents.tracing import Span, SpanError, agent_span, get_current_trace, trace
from agents.tracing.span_data import AgentSpanData
from agents.usage import Usage
from agents.util import _coro, _error_tracing
from agents.util._types import MaybeAwaitable

from typing_extensions import Unpack
import asyncio
from agents.tracing import (
    get_current_trace,
)
from dataclasses import dataclass, field


@dataclass
class ModelInputData:
    """Container for the data that will be sent to the model."""

    input: list[TResponseInputItem]
    instructions: str | None

@dataclass
class CallModelData(Generic[TContext]):
    """Data passed to `RunConfig.call_model_input_filter` prior to model call."""

    model_data: ModelInputData
    agent: Agent[TContext]
    context: TContext | None

class Runner:
    @classmethod
    async def run(
        cls,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        *,
        context: TContext | None = None,
        max_turns: int = DEFAULT_MAX_TURNS,
        hooks: RunHooks[TContext] | None = None,
        run_config: RunConfig | None = None,
        previous_response_id: str | None = None,
        session: Session | None = None,
    ) -> RunResult:
        """Run a workflow starting at the given agent. The agent will run in a loop until a final
        output is generated. The loop runs like so:
        1. The agent is invoked with the given input.
        2. If there is a final output (i.e. the agent produces something of type
            `agent.output_type`, the loop terminates.
        3. If there's a handoff, we run the loop again, with the new agent.
        4. Else, we run tool calls (if any), and re-run the loop.
        In two cases, the agent may raise an exception:
        1. If the max_turns is exceeded, a MaxTurnsExceeded exception is raised.
        2. If a guardrail tripwire is triggered, a GuardrailTripwireTriggered exception is raised.
        Note that only the first agent's input guardrails are run.
        Args:
            starting_agent: The starting agent to run.
            input: The initial input to the agent. You can pass a single string for a user message,
                or a list of input items.
            context: The context to run the agent with.
            max_turns: The maximum number of turns to run the agent for. A turn is defined as one
                AI invocation (including any tool calls that might occur).
            hooks: An object that receives callbacks on various lifecycle events.
            run_config: Global settings for the entire agent run.
            previous_response_id: The ID of the previous response, if using OpenAI models via the
                Responses API, this allows you to skip passing in input from the previous turn.
        Returns:
            A run result containing all the inputs, guardrail results and the output of the last
            agent. Agents may perform handoffs, so we don't know the specific type of the output.
        """
        runner = DEFAULT_AGENT_RUNNER
        return await runner.run(
            starting_agent,
            input,
            context=context,
            max_turns=max_turns,
            hooks=hooks,
            run_config=run_config,
            previous_response_id=previous_response_id,
            session=session,
        )

    @classmethod
    def run_sync(
        cls,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        *,
        context: TContext | None = None,
        max_turns: int = DEFAULT_MAX_TURNS,
        hooks: RunHooks[TContext] | None = None,
        run_config: RunConfig | None = None,
        previous_response_id: str | None = None,
        session: Session | None = None,
    ) -> RunResult:
        """Run a workflow synchronously, starting at the given agent. Note that this just wraps the
        `run` method, so it will not work if there's already an event loop (e.g. inside an async
        function, or in a Jupyter notebook or async context like FastAPI). For those cases, use
        the `run` method instead.
        The agent will run in a loop until a final output is generated. The loop runs like so:
        1. The agent is invoked with the given input.
        2. If there is a final output (i.e. the agent produces something of type
            `agent.output_type`, the loop terminates.
        3. If there's a handoff, we run the loop again, with the new agent.
        4. Else, we run tool calls (if any), and re-run the loop.
        In two cases, the agent may raise an exception:
        1. If the max_turns is exceeded, a MaxTurnsExceeded exception is raised.
        2. If a guardrail tripwire is triggered, a GuardrailTripwireTriggered exception is raised.
        Note that only the first agent's input guardrails are run.
        Args:
            starting_agent: The starting agent to run.
            input: The initial input to the agent. You can pass a single string for a user message,
                or a list of input items.
            context: The context to run the agent with.
            max_turns: The maximum number of turns to run the agent for. A turn is defined as one
                AI invocation (including any tool calls that might occur).
            hooks: An object that receives callbacks on various lifecycle events.
            run_config: Global settings for the entire agent run.
            previous_response_id: The ID of the previous response, if using OpenAI models via the
                Responses API, this allows you to skip passing in input from the previous turn.
        Returns:
            A run result containing all the inputs, guardrail results and the output of the last
            agent. Agents may perform handoffs, so we don't know the specific type of the output.
        """
        runner = DEFAULT_AGENT_RUNNER
        return runner.run_sync(
            starting_agent,
            input,
            context=context,
            max_turns=max_turns,
            hooks=hooks,
            run_config=run_config,
            previous_response_id=previous_response_id,
            session=session,
        )

    @classmethod
    def run_streamed(
        cls,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        context: TContext | None = None,
        max_turns: int = DEFAULT_MAX_TURNS,
        hooks: RunHooks[TContext] | None = None,
        run_config: RunConfig | None = None,
        previous_response_id: str | None = None,
        session: Session | None = None,
    ) -> RunResultStreaming:
        """Run a workflow starting at the given agent in streaming mode. The returned result object
        contains a method you can use to stream semantic events as they are generated.
        The agent will run in a loop until a final output is generated. The loop runs like so:
        1. The agent is invoked with the given input.
        2. If there is a final output (i.e. the agent produces something of type
            `agent.output_type`, the loop terminates.
        3. If there's a handoff, we run the loop again, with the new agent.
        4. Else, we run tool calls (if any), and re-run the loop.
        In two cases, the agent may raise an exception:
        1. If the max_turns is exceeded, a MaxTurnsExceeded exception is raised.
        2. If a guardrail tripwire is triggered, a GuardrailTripwireTriggered exception is raised.
        Note that only the first agent's input guardrails are run.
        Args:
            starting_agent: The starting agent to run.
            input: The initial input to the agent. You can pass a single string for a user message,
                or a list of input items.
            context: The context to run the agent with.
            max_turns: The maximum number of turns to run the agent for. A turn is defined as one
                AI invocation (including any tool calls that might occur).
            hooks: An object that receives callbacks on various lifecycle events.
            run_config: Global settings for the entire agent run.
            previous_response_id: The ID of the previous response, if using OpenAI models via the
                Responses API, this allows you to skip passing in input from the previous turn.
        Returns:
            A result object that contains data about the run, as well as a method to stream events.
        """
        runner = DEFAULT_AGENT_RUNNER
        return runner.run_streamed(
            starting_agent,
            input,
            context=context,
            max_turns=max_turns,
            hooks=hooks,
            run_config=run_config,
            previous_response_id=previous_response_id,
            session=session,
        )


class AgentRunner:
    async def run(
        self,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        **kwargs: Unpack[RunOptions[TContext]],
    ) -> RunResult:
        context = kwargs.get("context")
        max_turns = kwargs.get("max_turns", DEFAULT_MAX_TURNS)
        hooks = kwargs.get("hooks")
        run_config = kwargs.get("run_config")
        previous_response_id = kwargs.get("previous_response_id")
        session = kwargs.get("session")
        if hooks is None:
            hooks = RunHooks[Any]()
        if run_config is None:
            run_config = RunConfig()

        # Prepare input with session if enabled
        prepared_input = await self._prepare_input_with_session(input, session)

        tool_use_tracker = AgentToolUseTracker()

        with TraceCtxManager(
            workflow_name=run_config.workflow_name,
            trace_id=run_config.trace_id,
            group_id=run_config.group_id,
            metadata=run_config.trace_metadata,
            disabled=run_config.tracing_disabled,
        ):
            current_turn = 0
            original_input: str | list[TResponseInputItem] = _copy_str_or_list(prepared_input)
            generated_items: list[RunItem] = []
            model_responses: list[ModelResponse] = []

            context_wrapper: RunContextWrapper[TContext] = RunContextWrapper(
                context=context  # type: ignore
            )

            input_guardrail_results: list[InputGuardrailResult] = []

            current_span: Span[AgentSpanData] | None = None
            current_agent = starting_agent
            should_run_agent_start_hooks = True

            try:
                while True:
                    all_tools = await AgentRunner._get_all_tools(current_agent, context_wrapper)

                    # Start an agent span if we don't have one
                    if current_span is None:
                        handoff_names = [
                            h.agent_name
                            for h in await AgentRunner._get_handoffs(current_agent, context_wrapper)
                        ]
                        if output_schema := AgentRunner._get_output_schema(current_agent):
                            output_type_name = output_schema.name()
                        else:
                            output_type_name = "str"

                        current_span = agent_span(
                            name=current_agent.name,
                            handoffs=handoff_names,
                            output_type=output_type_name,
                        )
                        current_span.start(mark_as_current=True)
                        current_span.span_data.tools = [t.name for t in all_tools]

                    current_turn += 1
                    if current_turn > max_turns:
                        _error_tracing.attach_error_to_span(
                            current_span,
                            SpanError(
                                message="Max turns exceeded",
                                data={"max_turns": max_turns},
                            ),
                        )
                        raise MaxTurnsExceeded(f"Max turns ({max_turns}) exceeded")

                    # Run input guardrails for the first turn or after a handoff
                    if current_turn == 1 or isinstance(turn_result.next_step, NextStepHandoff):
                        input_guardrail_results.extend(
                            await self._run_input_guardrails(
                                current_agent,
                                current_agent.input_guardrails + (run_config.input_guardrails or []),
                                _copy_str_or_list(prepared_input if current_turn == 1 else original_input),
                                context_wrapper,
                            )
                        )
                        # Update original_input with sanitized content if any guardrail used 'retry'
                        for result in input_guardrail_results:
                            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                                original_input = result.output.sanitized_content

                    turn_result = await self._run_single_turn(
                        agent=current_agent,
                        all_tools=all_tools,
                        original_input=original_input,
                        generated_items=generated_items,
                        hooks=hooks,
                        context_wrapper=context_wrapper,
                        run_config=run_config,
                        should_run_agent_start_hooks=should_run_agent_start_hooks,
                        tool_use_tracker=tool_use_tracker,
                        previous_response_id=previous_response_id,
                    )
                    should_run_agent_start_hooks = False

                    model_responses.append(turn_result.model_response)
                    original_input = turn_result.original_input
                    generated_items = turn_result.generated_items

                    if isinstance(turn_result.next_step, NextStepFinalOutput):
                        output_guardrail_results = await self._run_output_guardrails(
                            current_agent.output_guardrails + (run_config.output_guardrails or []),
                            current_agent,
                            turn_result.next_step.output,
                            context_wrapper,
                        )
                        # Use sanitized content as final output if available
                        final_output = turn_result.next_step.output
                        for result in output_guardrail_results:
                            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                                final_output = result.output.sanitized_content

                        result = RunResult(
                            input=original_input,
                            new_items=generated_items,
                            raw_responses=model_responses,
                            final_output=final_output,
                            _last_agent=current_agent,
                            input_guardrail_results=input_guardrail_results,
                            output_guardrail_results=output_guardrail_results,
                            context_wrapper=context_wrapper,
                        )

                        # Save the conversation to session if enabled
                        await self._save_result_to_session(session, input, result)

                        return result
                    elif isinstance(turn_result.next_step, NextStepHandoff):
                        current_agent = cast(Agent[TContext], turn_result.next_step.new_agent)
                        # Pass sanitized input to the handoff agent
                        for result in input_guardrail_results:
                            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                                original_input = result.output.sanitized_content
                        current_span.finish(reset_current=True)
                        current_span = None
                        should_run_agent_start_hooks = True
                    elif isinstance(turn_result.next_step, NextStepRunAgain):
                        pass
                    else:
                        raise AgentsException(
                            f"Unknown next step type: {type(turn_result.next_step)}"
                        )
            except AgentsException as exc:
                exc.run_data = RunErrorDetails(
                    input=original_input,
                    new_items=generated_items,
                    raw_responses=model_responses,
                    last_agent=current_agent,
                    context_wrapper=context_wrapper,
                    input_guardrail_results=input_guardrail_results,
                    output_guardrail_results=[],
                )
                raise
            finally:
                if current_span:
                    current_span.finish(reset_current=True)

    def run_sync(
        self,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        **kwargs: Unpack[RunOptions[TContext]],
    ) -> RunResult:
        context = kwargs.get("context")
        max_turns = kwargs.get("max_turns", DEFAULT_MAX_TURNS)
        hooks = kwargs.get("hooks")
        run_config = kwargs.get("run_config")
        previous_response_id = kwargs.get("previous_response_id")
        session = kwargs.get("session")

        return asyncio.get_event_loop().run_until_complete(
            self.run(
                starting_agent,
                input,
                session=session,
                context=context,
                max_turns=max_turns,
                hooks=hooks,
                run_config=run_config,
                previous_response_id=previous_response_id,
            )
        )

    def run_streamed(
        self,
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        **kwargs: Unpack[RunOptions[TContext]],
    ) -> RunResultStreaming:
        context = kwargs.get("context")
        max_turns = kwargs.get("max_turns", DEFAULT_MAX_TURNS)
        hooks = kwargs.get("hooks")
        run_config = kwargs.get("run_config")
        previous_response_id = kwargs.get("previous_response_id")
        session = kwargs.get("session")

        if hooks is None:
            hooks = RunHooks[Any]()
        if run_config is None:
            run_config = RunConfig()

        new_trace = (
            None
            if get_current_trace()
            else trace(
                workflow_name=run_config.workflow_name,
                trace_id=run_config.trace_id,
                group_id=run_config.group_id,
                metadata=run_config.trace_metadata,
                disabled=run_config.tracing_disabled,
            )
        )

        output_schema = AgentRunner._get_output_schema(starting_agent)
        context_wrapper: RunContextWrapper[TContext] = RunContextWrapper(
            context=context  # type: ignore
        )

        streamed_result = RunResultStreaming(
            input=_copy_str_or_list(input),
            new_items=[],
            current_agent=starting_agent,
            raw_responses=[],
            final_output=None,
            is_complete=False,
            current_turn=0,
            max_turns=max_turns,
            input_guardrail_results=[],
            output_guardrail_results=[],
            _current_agent_output_schema=output_schema,
            trace=new_trace,
            context_wrapper=context_wrapper,
        )

        streamed_result._run_impl_task = asyncio.create_task(
            self._start_streaming(
                starting_input=input,
                streamed_result=streamed_result,
                starting_agent=starting_agent,
                max_turns=max_turns,
                hooks=hooks,
                context_wrapper=context_wrapper,
                run_config=run_config,
                previous_response_id=previous_response_id,
                session=session,
            )
        )
        return streamed_result

    @classmethod
    async def _maybe_filter_model_input(
        cls,
        *,
        agent: Agent[TContext],
        run_config: RunConfig,
        context_wrapper: RunContextWrapper[TContext],
        input_items: list[TResponseInputItem],
        system_instructions: str | None,
    ) -> ModelInputData:
        """Apply optional call_model_input_filter to modify model input.

        Returns a `ModelInputData` that will be sent to the model.
        """
        effective_instructions = system_instructions
        effective_input: list[TResponseInputItem] = input_items

        if run_config.call_model_input_filter is None:
            return ModelInputData(input=effective_input, instructions=effective_instructions)

        try:
            model_input = ModelInputData(
                input=effective_input.copy(),
                instructions=effective_instructions,
            )
            filter_payload: CallModelData[TContext] = CallModelData(
                model_data=model_input,
                agent=agent,
                context=context_wrapper.context,
            )
            maybe_updated = run_config.call_model_input_filter(filter_payload)
            updated = await maybe_updated if inspect.isawaitable(maybe_updated) else maybe_updated
            if not isinstance(updated, ModelInputData):
                raise UserError("call_model_input_filter must return a ModelInputData instance")
            return updated
        except Exception as e:
            _error_tracing.attach_error_to_current_span(
                SpanError(message="Error in call_model_input_filter", data={"error": str(e)})
            )
            raise

    @classmethod
    async def _run_input_guardrails_with_queue(
        cls,
        agent: Agent[Any],
        guardrails: list[InputGuardrail[TContext]],
        input: str | list[TResponseInputItem],
        context: RunContextWrapper[TContext],
        streamed_result: RunResultStreaming,
        parent_span: Span[Any],
    ):
        queue = streamed_result._input_guardrail_queue

        # We'll run the guardrails and push them onto the queue as they complete
        guardrail_tasks = [
            asyncio.create_task(
                RunImpl.run_single_input_guardrail(agent, guardrail, input, context)
            )
            for guardrail in guardrails
        ]
        guardrail_results = []
        try:
            for done in asyncio.as_completed(guardrail_tasks):
                result = await done
                if result.output.tripwire_triggered:
                    _error_tracing.attach_error_to_span(
                        parent_span,
                        SpanError(
                            message="Guardrail tripwire triggered",
                            data={
                                "guardrail": result.guardrail.get_name(),
                                "type": "input_guardrail",
                            },
                        ),
                    )
                queue.put_nowait(result)
                guardrail_results.append(result)
        except Exception:
            for t in guardrail_tasks:
                t.cancel()
            raise

        streamed_result.input_guardrail_results = guardrail_results
        # Update streamed_result.input with sanitized content if available
        for result in guardrail_results:
            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                streamed_result.input = result.output.sanitized_content

    @classmethod
    async def _start_streaming(
        cls,
        starting_input: str | list[TResponseInputItem],
        streamed_result: RunResultStreaming,
        starting_agent: Agent[TContext],
        max_turns: int,
        hooks: RunHooks[TContext],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        previous_response_id: str | None,
        session: Session | None,
    ):
        if streamed_result.trace:
            streamed_result.trace.start(mark_as_current=True)

        current_span: Span[AgentSpanData] | None = None
        current_agent = starting_agent
        current_turn = 0
        should_run_agent_start_hooks = True
        tool_use_tracker = AgentToolUseTracker()

        streamed_result._event_queue.put_nowait(AgentUpdatedStreamEvent(new_agent=current_agent))

        try:
            # Prepare input with session if enabled
            prepared_input = await AgentRunner._prepare_input_with_session(starting_input, session)

            # Update the streamed result with the prepared input
            streamed_result.input = prepared_input

            while True:
                if streamed_result.is_complete:
                    break

                all_tools = await cls._get_all_tools(current_agent, context_wrapper)

                # Start an agent span if we don't have one
                if current_span is None:
                    handoff_names = [
                        h.agent_name
                        for h in await cls._get_handoffs(current_agent, context_wrapper)
                    ]
                    if output_schema := cls._get_output_schema(current_agent):
                        output_type_name = output_schema.name()
                    else:
                        output_type_name = "str"

                    current_span = agent_span(
                        name=current_agent.name,
                        handoffs=handoff_names,
                        output_type=output_type_name,
                    )
                    current_span.start(mark_as_current=True)
                    tool_names = [t.name for t in all_tools]
                    current_span.span_data.tools = tool_names
                current_turn += 1
                streamed_result.current_turn = current_turn

                if current_turn > max_turns:
                    _error_tracing.attach_error_to_span(
                        current_span,
                        SpanError(
                            message="Max turns exceeded",
                            data={"max_turns": max_turns},
                        ),
                    )
                    streamed_result._event_queue.put_nowait(QueueCompleteSentinel())
                    break

                if current_turn == 1 or isinstance(turn_result.next_step, NextStepHandoff):
                    # Run input guardrails in the background and put the results on the queue
                    streamed_result._input_guardrails_task = asyncio.create_task(
                        cls._run_input_guardrails_with_queue(
                            current_agent,
                            current_agent.input_guardrails + (run_config.input_guardrails or []),
                            ItemHelpers.input_to_new_input_list(prepared_input if current_turn == 1 else streamed_result.input),
                            context_wrapper,
                            streamed_result,
                            current_span,
                        )
                    )
                try:
                    turn_result = await cls._run_single_turn_streamed(
                        streamed_result,
                        current_agent,
                        hooks,
                        context_wrapper,
                        run_config,
                        should_run_agent_start_hooks,
                        tool_use_tracker,
                        all_tools,
                        previous_response_id,
                    )
                    should_run_agent_start_hooks = False

                    streamed_result.raw_responses = streamed_result.raw_responses + [
                        turn_result.model_response
                    ]
                    streamed_result.input = turn_result.original_input
                    streamed_result.new_items = turn_result.generated_items

                    if isinstance(turn_result.next_step, NextStepHandoff):
                        current_agent = turn_result.next_step.new_agent
                        # Update input with sanitized content from input guardrails for handoff
                        for result in streamed_result.input_guardrail_results:
                            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                                streamed_result.input = result.output.sanitized_content
                        current_span.finish(reset_current=True)
                        current_span = None
                        should_run_agent_start_hooks = True
                        streamed_result._event_queue.put_nowait(
                            AgentUpdatedStreamEvent(new_agent=current_agent)
                        )
                    elif isinstance(turn_result.next_step, NextStepFinalOutput):
                        streamed_result._output_guardrails_task = asyncio.create_task(
                            cls._run_output_guardrails(
                                current_agent.output_guardrails
                                + (run_config.output_guardrails or []),
                                current_agent,
                                turn_result.next_step.output,
                                context_wrapper,
                            )
                        )

                        try:
                            output_guardrail_results = await streamed_result._output_guardrails_task
                        except Exception:
                            output_guardrail_results = []

                        streamed_result.output_guardrail_results = output_guardrail_results
                        # Use sanitized content as final output if available
                        final_output = turn_result.next_step.output
                        for result in output_guardrail_results:
                            if hasattr(result.output, 'sanitized_content') and result.output.sanitized_content:
                                final_output = result.output.sanitized_content
                        streamed_result.final_output = final_output
                        streamed_result.is_complete = True

                        # Save the conversation to session if enabled
                        temp_result = RunResult(
                            input=streamed_result.input,
                            new_items=streamed_result.new_items,
                            raw_responses=streamed_result.raw_responses,
                            final_output=streamed_result.final_output,
                            _last_agent=current_agent,
                            input_guardrail_results=streamed_result.input_guardrail_results,
                            output_guardrail_results=streamed_result.output_guardrail_results,
                            context_wrapper=context_wrapper,
                        )
                        await AgentRunner._save_result_to_session(
                            session, starting_input, temp_result
                        )

                        streamed_result._event_queue.put_nowait(QueueCompleteSentinel())
                    elif isinstance(turn_result.next_step, NextStepRunAgain):
                        pass
                except AgentsException as exc:
                    streamed_result.is_complete = True
                    streamed_result._event_queue.put_nowait(QueueCompleteSentinel())
                    exc.run_data = RunErrorDetails(
                        input=streamed_result.input,
                        new_items=streamed_result.new_items,
                        raw_responses=streamed_result.raw_responses,
                        last_agent=current_agent,
                        context_wrapper=context_wrapper,
                        input_guardrail_results=streamed_result.input_guardrail_results,
                        output_guardrail_results=streamed_result.output_guardrail_results,
                    )
                    raise
                except Exception as e:
                    if current_span:
                        _error_tracing.attach_error_to_span(
                            current_span,
                            SpanError(
                                message="Error in agent run",
                                data={"error": str(e)},
                            ),
                        )
                    streamed_result.is_complete = True
                    streamed_result._event_queue.put_nowait(QueueCompleteSentinel())
                    raise

            streamed_result.is_complete = True
        finally:
            if current_span:
                current_span.finish(reset_current=True)
            if streamed_result.trace:
                streamed_result.trace.finish(reset_current=True)

    @classmethod
    async def _run_single_turn_streamed(
        cls,
        streamed_result: RunResultStreaming,
        agent: Agent[TContext],
        hooks: RunHooks[TContext],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        should_run_agent_start_hooks: bool,
        tool_use_tracker: AgentToolUseTracker,
        all_tools: list[Tool],
        previous_response_id: str | None,
    ) -> SingleStepResult:
        if should_run_agent_start_hooks:
            await asyncio.gather(
                hooks.on_agent_start(context_wrapper, agent),
                (
                    agent.hooks.on_start(context_wrapper, agent)
                    if agent.hooks
                    else _coro.noop_coroutine()
                ),
            )

        output_schema = cls._get_output_schema(agent)

        streamed_result.current_agent = agent
        streamed_result._current_agent_output_schema = output_schema

        system_prompt, prompt_config = await asyncio.gather(
            agent.get_system_prompt(context_wrapper),
            agent.get_prompt(context_wrapper),
        )

        handoffs = await cls._get_handoffs(agent, context_wrapper)
        model = cls._get_model(agent, run_config)
        model_settings = agent.model_settings.resolve(run_config.model_settings)
        model_settings = RunImpl.maybe_reset_tool_choice(agent, tool_use_tracker, model_settings)

        final_response: ModelResponse | None = None

        input = ItemHelpers.input_to_new_input_list(streamed_result.input)
        input.extend([item.to_input_item() for item in streamed_result.new_items])

        filtered = await cls._maybe_filter_model_input(
            agent=agent,
            run_config=run_config,
            context_wrapper=context_wrapper,
            input_items=input,
            system_instructions=system_prompt,
        )

        # Call hook just before the model is invoked, with the correct system_prompt.
        if agent.hooks:
            await agent.hooks.on_llm_start(
                context_wrapper, agent, filtered.instructions, filtered.input
            )

        # 1. Stream the output events
        async for event in model.stream_response(
            filtered.instructions,
            filtered.input,
            model_settings,
            all_tools,
            output_schema,
            handoffs,
            get_model_tracing_impl(
                run_config.tracing_disabled, run_config.trace_include_sensitive_data
            ),
            previous_response_id=previous_response_id,
            prompt=prompt_config,
        ):
            if isinstance(event, ResponseCompletedEvent):
                usage = (
                    Usage(
                        requests=1,
                        input_tokens=event.response.usage.input_tokens,
                        output_tokens=event.response.usage.output_tokens,
                        total_tokens=event.response.usage.total_tokens,
                        input_tokens_details=event.response.usage.input_tokens_details,
                        output_tokens_details=event.response.usage.output_tokens_details,
                    )
                    if event.response.usage
                    else Usage()
                )
                final_response = ModelResponse(
                    output=event.response.output,
                    usage=usage,
                    response_id=event.response.id,
                )
                context_wrapper.usage.add(usage)

            streamed_result._event_queue.put_nowait(RawResponsesStreamEvent(data=event))

        # Call hook just after the model response is finalized.
        if agent.hooks and final_response is not None:
            await agent.hooks.on_llm_end(context_wrapper, agent, final_response)

        # 2. At this point, the streaming is complete for this turn of the agent loop.
        if not final_response:
            raise ModelBehaviorError("Model did not produce a final response!")

        # 3. Now, we can process the turn as we do in the non-streaming case
        return await cls._get_single_step_result_from_streamed_response(
            agent=agent,
            streamed_result=streamed_result,
            new_response=final_response,
            output_schema=output_schema,
            all_tools=all_tools,
            handoffs=handoffs,
            hooks=hooks,
            context_wrapper=context_wrapper,
            run_config=run_config,
            tool_use_tracker=tool_use_tracker,
        )

    @classmethod
    async def _run_single_turn(
        cls,
        *,
        agent: Agent[TContext],
        all_tools: list[Tool],
        original_input: str | list[TResponseInputItem],
        generated_items: list[RunItem],
        hooks: RunHooks[TContext],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        should_run_agent_start_hooks: bool,
        tool_use_tracker: AgentToolUseTracker,
        previous_response_id: str | None,
    ) -> SingleStepResult:
        # Ensure we run the hooks before anything else
        if should_run_agent_start_hooks:
            await asyncio.gather(
                hooks.on_agent_start(context_wrapper, agent),
                (
                    agent.hooks.on_start(context_wrapper, agent)
                    if agent.hooks
                    else _coro.noop_coroutine()
                ),
            )

        system_prompt, prompt_config = await asyncio.gather(
            agent.get_system_prompt(context_wrapper),
            agent.get_prompt(context_wrapper),
        )

        output_schema = cls._get_output_schema(agent)
        handoffs = await cls._get_handoffs(agent, context_wrapper)
        input = ItemHelpers.input_to_new_input_list(original_input)
        input.extend([generated_item.to_input_item() for generated_item in generated_items])

        new_response = await cls._get_new_response(
            agent,
            system_prompt,
            input,
            output_schema,
            all_tools,
            handoffs,
            context_wrapper,
            run_config,
            tool_use_tracker,
            previous_response_id,
            prompt_config,
        )

        return await cls._get_single_step_result_from_response(
            agent=agent,
            original_input=original_input,
            pre_step_items=generated_items,
            new_response=new_response,
            output_schema=output_schema,
            all_tools=all_tools,
            handoffs=handoffs,
            hooks=hooks,
            context_wrapper=context_wrapper,
            run_config=run_config,
            tool_use_tracker=tool_use_tracker,
        )

    @classmethod
    async def _get_single_step_result_from_response(
        cls,
        *,
        agent: Agent[TContext],
        all_tools: list[Tool],
        original_input: str | list[TResponseInputItem],
        pre_step_items: list[RunItem],
        new_response: ModelResponse,
        output_schema: AgentOutputSchemaBase | None,
        handoffs: list[Handoff],
        hooks: RunHooks[TContext],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        tool_use_tracker: AgentToolUseTracker,
    ) -> SingleStepResult:
        processed_response = RunImpl.process_model_response(
            agent=agent,
            all_tools=all_tools,
            response=new_response,
            output_schema=output_schema,
            handoffs=handoffs,
        )

        tool_use_tracker.add_tool_use(agent, processed_response.tools_used)

        return await RunImpl.execute_tools_and_side_effects(
            agent=agent,
            original_input=original_input,
            pre_step_items=pre_step_items,
            new_response=new_response,
            processed_response=processed_response,
            output_schema=output_schema,
            hooks=hooks,
            context_wrapper=context_wrapper,
            run_config=run_config,
        )

    @classmethod
    async def _get_single_step_result_from_streamed_response(
        cls,
        *,
        agent: Agent[TContext],
        all_tools: list[Tool],
        streamed_result: RunResultStreaming,
        new_response: ModelResponse,
        output_schema: AgentOutputSchemaBase | None,
        handoffs: list[Handoff],
        hooks: RunHooks[TContext],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        tool_use_tracker: AgentToolUseTracker,
    ) -> SingleStepResult:
        original_input = streamed_result.input
        pre_step_items = streamed_result.new_items
        event_queue = streamed_result._event_queue

        processed_response = RunImpl.process_model_response(
            agent=agent,
            all_tools=all_tools,
            response=new_response,
            output_schema=output_schema,
            handoffs=handoffs,
        )
        new_items_processed_response = processed_response.new_items
        tool_use_tracker.add_tool_use(agent, processed_response.tools_used)
        RunImpl.stream_step_items_to_queue(new_items_processed_response, event_queue)

        single_step_result = await RunImpl.execute_tools_and_side_effects(
            agent=agent,
            original_input=original_input,
            pre_step_items=pre_step_items,
            new_response=new_response,
            processed_response=processed_response,
            output_schema=output_schema,
            hooks=hooks,
            context_wrapper=context_wrapper,
            run_config=run_config,
        )
        new_step_items = [
            item
            for item in single_step_result.new_step_items
            if item not in new_items_processed_response
        ]
        RunImpl.stream_step_items_to_queue(new_step_items, event_queue)

        return single_step_result

    @classmethod
    async def _run_input_guardrails(
        cls,
        agent: Agent[Any],
        guardrails: list[InputGuardrail[TContext]],
        input: str | list[TResponseInputItem],
        context: RunContextWrapper[TContext],
    ) -> list[InputGuardrailResult]:
        if not guardrails:
            return []

        guardrail_tasks = [
            asyncio.create_task(
                RunImpl.run_single_input_guardrail(agent, guardrail, input, context)
            )
            for guardrail in guardrails
        ]

        guardrail_results = []

        for done in asyncio.as_completed(guardrail_tasks):
            result = await done
            if result.output.tripwire_triggered:
                # Cancel all guardrail tasks if a tripwire is triggered.
                for t in guardrail_tasks:
                    t.cancel()
                _error_tracing.attach_error_to_current_span(
                    SpanError(
                        message="Guardrail tripwire triggered",
                        data={"guardrail": result.guardrail.get_name()},
                    )
                )
                raise InputGuardrailTripwireTriggered(result)
            else:
                guardrail_results.append(result)

        return guardrail_results

    @classmethod
    async def _run_output_guardrails(
        cls,
        guardrails: list[OutputGuardrail[TContext]],
        agent: Agent[TContext],
        agent_output: Any,
        context: RunContextWrapper[TContext],
    ) -> list[OutputGuardrailResult]:
        if not guardrails:
            return []

        guardrail_tasks = [
            asyncio.create_task(
                RunImpl.run_single_output_guardrail(guardrail, agent, agent_output, context)
            )
            for guardrail in guardrails
        ]

        guardrail_results = []

        for done in asyncio.as_completed(guardrail_tasks):
            result = await done
            if result.output.tripwire_triggered:
                # Cancel all guardrail tasks if a tripwire is triggered.
                for t in guardrail_tasks:
                    t.cancel()
                _error_tracing.attach_error_to_current_span(
                    SpanError(
                        message="Guardrail tripwire triggered",
                        data={"guardrail": result.guardrail.get_name()},
                    )
                )
                raise OutputGuardrailTripwireTriggered(result)
            else:
                guardrail_results.append(result)

        return guardrail_results

    @classmethod
    async def _get_new_response(
        cls,
        agent: Agent[TContext],
        system_prompt: str | None,
        input: list[TResponseInputItem],
        output_schema: AgentOutputSchemaBase | None,
        all_tools: list[Tool],
        handoffs: list[Handoff],
        context_wrapper: RunContextWrapper[TContext],
        run_config: RunConfig,
        tool_use_tracker: AgentToolUseTracker,
        previous_response_id: str | None,
        prompt_config: ResponsePromptParam | None,
    ) -> ModelResponse:
        # Allow user to modify model input right before the call, if configured
        filtered = await cls._maybe_filter_model_input(
            agent=agent,
            run_config=run_config,
            context_wrapper=context_wrapper,
            input_items=input,
            system_instructions=system_prompt,
        )

        model = cls._get_model(agent, run_config)
        model_settings = agent.model_settings.resolve(run_config.model_settings)
        model_settings = RunImpl.maybe_reset_tool_choice(agent, tool_use_tracker, model_settings)
        # If the agent has hooks, we need to call them before and after the LLM call
        if agent.hooks:
            await agent.hooks.on_llm_start(
                context_wrapper,
                agent,
                filtered.instructions,  # Use filtered instructions
                filtered.input,  # Use filtered input
            )

        new_response = await model.get_response(
            system_instructions=filtered.instructions,
            input=filtered.input,
            model_settings=model_settings,
            tools=all_tools,
            output_schema=output_schema,
            handoffs=handoffs,
            tracing=get_model_tracing_impl(
                run_config.tracing_disabled, run_config.trace_include_sensitive_data
            ),
            previous_response_id=previous_response_id,
            prompt=prompt_config,
        )
        # If the agent has hooks, we need to call them after the LLM call
        if agent.hooks:
            await agent.hooks.on_llm_end(context_wrapper, agent, new_response)

        context_wrapper.usage.add(new_response.usage)

        return new_response

    @classmethod
    def _get_output_schema(cls, agent: Agent[Any]) -> AgentOutputSchemaBase | None:
        if agent.output_type is None or agent.output_type is str:
            return None
        elif isinstance(agent.output_type, AgentOutputSchemaBase):
            return agent.output_type

        return AgentOutputSchema(agent.output_type)

    @classmethod
    async def _get_handoffs(
        cls, agent: Agent[Any], context_wrapper: RunContextWrapper[Any]
    ) -> list[Handoff]:
        handoffs = []
        for handoff_item in agent.handoffs:
            if isinstance(handoff_item, Handoff):
                handoffs.append(handoff_item)
            elif isinstance(handoff_item, Agent):
                handoffs.append(handoff(handoff_item))

        async def _check_handoff_enabled(handoff_obj: Handoff) -> bool:
            attr = handoff_obj.is_enabled
            if isinstance(attr, bool):
                return attr
            res = attr(context_wrapper, agent)
            if inspect.isawaitable(res):
                return bool(await res)
            return bool(res)

        results = await asyncio.gather(*(_check_handoff_enabled(h) for h in handoffs))
        enabled: list[Handoff] = [h for h, ok in zip(handoffs, results) if ok]
        return enabled

    @classmethod
    async def _get_all_tools(
        cls, agent: Agent[Any], context_wrapper: RunContextWrapper[Any]
    ) -> list[Tool]:
        return await agent.get_all_tools(context_wrapper)

    @classmethod
    def _get_model(cls, agent: Agent[Any], run_config: RunConfig) -> Model:
        if isinstance(run_config.model, Model):
            return run_config.model
        elif isinstance(run_config.model, str):
            return run_config.model_provider.get_model(run_config.model)
        elif isinstance(agent.model, Model):
            return agent.model

        return run_config.model_provider.get_model(agent.model)

    @classmethod
    async def _prepare_input_with_session(
        cls,
        input: str | list[TResponseInputItem],
        session: Session | None,
    ) -> str | list[TResponseInputItem]:
        """Prepare input by combining it with session history if enabled."""
        if session is None:
            return input

        # Validate that we don't have both a session and a list input, as this creates
        # ambiguity about whether the list should append to or replace existing session history
        if isinstance(input, list):
            raise UserError(
                "Cannot provide both a session and a list of input items. "
                "When using session memory, provide only a string input to append to the "
                "conversation, or use session=None and provide a list to manually manage "
                "conversation history."
            )

        # Get previous conversation history
        history = await session.get_items()

        # Convert input to list format
        new_input_list = ItemHelpers.input_to_new_input_list(input)

        # Combine history with new input
        combined_input = history + new_input_list

        return combined_input

    @classmethod
    async def _save_result_to_session(
        cls,
        session: Session | None,
        original_input: str | list[TResponseInputItem],
        result: RunResult,
    ) -> None:
        """Save the conversation turn to session."""
        if session is None:
            return

        # Convert original input to list format if needed
        input_list = ItemHelpers.input_to_new_input_list(original_input)

        # Convert new items to input format
        new_items_as_input = [item.to_input_item() for item in result.new_items]

        # Save all items from this turn
        items_to_save = input_list + new_items_as_input
        await session.add_items(items_to_save)


DEFAULT_AGENT_RUNNER = AgentRunner()


def _copy_str_or_list(input: str | list[TResponseInputItem]) -> str | list[TResponseInputItem]:
    if isinstance(input, str):
        return input
    return input.copy()


import pandas as pd
from pydantic import BaseModel
from typing import Dict, List, Optional
from aryaxai.client.client import APIClient
from aryaxai.common.validation import Validate
from aryaxai.common.xai_uris import (
    AVAILABLE_CUSTOM_SERVERS_URI,
    CREATE_WORKSPACE_URI,
    GET_WORKSPACES_URI,
    INVITE_USER_ORGANIZATION_URI,
    ORGANIZATION_MEMBERS_URI,
    REMOVE_USER_ORGANIZATION_URI,
)
from aryaxai.core.workspace import Workspace
from aryaxai.common.types import GCSConfig, S3Config, GDriveConfig, SFTPConfig
from aryaxai.common.xai_uris import (
    AVAILABLE_CUSTOM_SERVERS_URI,
    CREATE_DATA_CONNECTORS,
    LIST_DATA_CONNECTORS,
    DELETE_DATA_CONNECTORS,
    TEST_DATA_CONNECTORS,
    DROPBOX_OAUTH,
    LIST_BUCKETS,
    LIST_FILEPATHS,
    COMPUTE_CREDIT_URI,
)
from aryaxai.core.utils import build_url, build_list_data_connector_url


class Organization(BaseModel):
    """Class to work with AryaXAI organizations"""

    organization_id: Optional[str] = None
    name: str
    created_by: str
    created_at: Optional[str] = None

    api_client: APIClient

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.api_client = kwargs.get("api_client")

    def add_user_to_organization(self, user_email: str) -> str:
        """Add user to Organization

        :param user_email: Email of user to be added to organization.
        :return: response
        """
        payload = {
            "email": user_email,
            "organization_id": self.organization_id,
        }
        res = self.api_client.post(INVITE_USER_ORGANIZATION_URI, payload)

        if not res["success"]:
            raise Exception(res.get("details", "Failed to add user to organization"))

        return res.get("details", "User added successfully")

    def remove_user_from_organization(self, user_email: str) -> str:
        """Remove user from Organization

        :param user_email: Email of user to be removed from organization.
        :return: response
        """
        payload = {
            "organization_user_email": user_email,
            "organization_id": self.organization_id,
        }
        res = self.api_client.post(REMOVE_USER_ORGANIZATION_URI, payload)

        if not res["success"]:
            raise Exception(
                res.get("details", "Failed to remove user from organization")
            )

        return res.get("details", "User removed successfully")

    def member_details(self) -> pd.DataFrame:
        """Organization Member details

        :return: member details dataframe
        """
        res = self.api_client.get(
            f"{ORGANIZATION_MEMBERS_URI}?organization_id={self.organization_id}"
        )

        if not res["success"]:
            raise Exception(
                res.get("details", "Failed to get organization member details")
            )

        member_details_df = pd.DataFrame(
            res.get("details").get("users"),
            columns=[
                "full_name",
                "email",
                "organization_owner",
                "organization_admin",
                "created_at",
            ],
        )

        return member_details_df

    def workspaces(self) -> pd.DataFrame:
        """get user workspaces

        :return: workspace details dataframe
        """

        url = GET_WORKSPACES_URI
        if self.organization_id:
            url = url + f"?organization_id={self.organization_id}"
        workspaces = self.api_client.get(url)

        workspace_df = pd.DataFrame(
            workspaces["details"],
            columns=[
                "user_workspace_name",
                "access_type",
                "created_by",
                "created_at",
                "updated_at",
                "instance_type",
                "instance_status",
            ],
        )

        return workspace_df

    def workspace(self, workspace_name: str) -> Workspace:
        """select specific workspace

        :param workspace_name: Name of the workspace to be used
        :return: Workspace
        """

        url = GET_WORKSPACES_URI
        if self.organization_id:
            url = url + f"?organization_id={self.organization_id}"
        workspaces = self.api_client.get(url)
        user_workspaces = [
            Workspace(api_client=self.api_client, **workspace)
            for workspace in workspaces["details"]
        ]

        workspace = next(
            filter(
                lambda workspace: workspace.user_workspace_name == workspace_name,
                user_workspaces,
            ),
            None,
        )

        if workspace is None:
            raise Exception("Workspace Not Found")

        return workspace

    def create_workspace(
        self, workspace_name: str, server_type: Optional[str] = None
    ) -> Workspace:
        """create user workspace

        :param workspace_name: name for the workspace
        :param server_type: dedicated instance to run workloads
            for all available instances check xai.available_custom_servers()
            defaults to shared
        :return: response
        """
        payload = {"workspace_name": workspace_name}

        if server_type:
            custom_servers = self.api_client.get(AVAILABLE_CUSTOM_SERVERS_URI)
            Validate.value_against_list(
                "server_type",
                server_type,
                [server["name"] for server in custom_servers],
            )

            payload["instance_type"] = server_type
            payload["server_config"] = {}

        if self.organization_id:
            payload["organization_id"] = self.organization_id

        res = self.api_client.post(CREATE_WORKSPACE_URI, payload)

        if not res["success"]:
            raise Exception(res.get("details"))

        workspace = Workspace(api_client=self.api_client, **res["workspace_details"])

        return workspace

    def __print__(self) -> str:
        return f"Organization(name='{self.name}', created_by='{self.created_by}', created_at='{self.created_at}')"

    def __str__(self) -> str:
        return self.__print__()

    def __repr__(self) -> str:
        return self.__print__()

    def create_data_connectors(
        self,
        data_connector_name: str,
        data_connector_type: str,
        gcs_config: Optional[GCSConfig] = None,
        s3_config: Optional[S3Config] = None,
        gdrive_config: Optional[GDriveConfig] = None,
        sftp_config: Optional[SFTPConfig] = None,
    ) -> str:
        """Create Data Connectors for project

        :param data_connector_name: str # name for data connector
        :param data_connector_type: str # type of data connector (s3 | gcs | gdrive)
        :param gcs_config: dict # credentials from service account json
        :param s3_config: dict # credentials of s3 storage
        :param gdrive_config: dict # credentials from service account json
        :param sftp_config: dict # hostname, port, username and password for sftp connection
        :return: response
        """
        if not self.organization_id:
            return "No Organization id found"
        if data_connector_type.lower() == "s3":
            if not s3_config:
                return "No configuration for S3 found"

            Validate.value_against_list(
                "s3 config",
                list(s3_config.keys()),
                ["region", "access_key", "secret_key"],
            )

            payload = {
                "link_service": {
                    "service_name": data_connector_name,
                    "region": s3_config.get("region", "ap-south-1"),
                    "access_key": s3_config.get("access_key"),
                    "secret_key": s3_config.get("secret_key"),
                },
                "link_service_type": data_connector_type,
            }

        if data_connector_type.lower() == "gcs":
            if not gcs_config:
                return "No configuration for GCS found"

            Validate.value_against_list(
                "gcs config",
                list(gcs_config.keys()),
                [
                    "project_id",
                    "gcp_project_name",
                    "type",
                    "private_key_id",
                    "private_key",
                    "client_email",
                    "client_id",
                    "auth_uri",
                    "token_uri",
                ],
            )

            payload = {
                "link_service": {
                    "service_name": data_connector_name,
                    "project_id": gcs_config.get("project_id"),
                    "gcp_project_name": gcs_config.get("gcp_project_name"),
                    "service_account_json": {
                        "type": gcs_config.get("type"),
                        "project_id": gcs_config.get("project_id"),
                        "private_key_id": gcs_config.get("private_key_id"),
                        "private_key": gcs_config.get("private_key"),
                        "client_email": gcs_config.get("client_email"),
                        "client_id": gcs_config.get("client_id"),
                        "auth_uri": gcs_config.get("auth_uri"),
                        "token_uri": gcs_config.get("token_uri"),
                    },
                },
                "link_service_type": data_connector_type,
            }

        if data_connector_type == "gdrive":
            if not gdrive_config:
                return "No configuration for Google Drive found"

            Validate.value_against_list(
                "gdrive config",
                list(gdrive_config.keys()),
                [
                    "project_id",
                    "type",
                    "private_key_id",
                    "private_key",
                    "client_email",
                    "client_id",
                    "auth_uri",
                    "token_uri",
                ],
            )

            payload = {
                "link_service": {
                    "service_name": data_connector_name,
                    "service_account_json": {
                        "type": gdrive_config.get("type"),
                        "project_id": gdrive_config.get("project_id"),
                        "private_key_id": gdrive_config.get("private_key_id"),
                        "private_key": gdrive_config.get("private_key"),
                        "client_email": gdrive_config.get("client_email"),
                        "client_id": gdrive_config.get("client_id"),
                        "auth_uri": gdrive_config.get("auth_uri"),
                        "token_uri": gdrive_config.get("token_uri"),
                    },
                },
                "link_service_type": data_connector_type,
            }

        if data_connector_type == "sftp":
            if not sftp_config:
                return "No configuration for Google Drive found"

            Validate.value_against_list(
                "sftp config",
                list(sftp_config.keys()),
                ["hostname", "port", "username", "password"],
            )

            payload = {
                "link_service": {
                    "service_name": data_connector_name,
                    "sftp_json": {
                        "hostname": sftp_config.get("hostname"),
                        "port": sftp_config.get("port"),
                        "username": sftp_config.get("username"),
                        "password": sftp_config.get("password"),
                    },
                },
                "link_service_type": data_connector_type,
            }

        if data_connector_type == "dropbox":
            url_data = self.api_client.get(
                f"{DROPBOX_OAUTH}?organization_id={self.organization_id}"
            )
            print(f"Url: {url_data['details']['url']}")
            code = input(f"{url_data['details']['message']}: ")

            if not code:
                return "No authentication code provided."

            payload = {
                "link_service": {
                    "service_name": data_connector_name,
                    "dropbox_json": {"code": code},
                },
                "link_service_type": data_connector_type,
            }

        url = build_url(
            CREATE_DATA_CONNECTORS, data_connector_name, None, self.organization_id
        )
        res = self.api_client.post(url, payload)
        return res["details"]

    def test_data_connectors(self, data_connector_name) -> str:
        """Test connection for the data connectors

        :param data_connector_name: str
        """
        if not data_connector_name:
            return "Missing argument data_connector_name"
        if not self.organization_id:
            return "No Project Name or Organization id found"
        url = build_url(
            TEST_DATA_CONNECTORS, data_connector_name, None, self.organization_id
        )
        res = self.api_client.post(url)
        return res["details"]

    def delete_data_connectors(self, data_connector_name) -> str:
        """Delete the data connectors

        :param data_connector_name: str
        """
        if not data_connector_name:
            return "Missing argument data_connector_name"
        if not self.organization_id:
            return "No Project Name or Organization id found"

        url = build_url(
            DELETE_DATA_CONNECTORS, data_connector_name, None, self.organization_id
        )
        res = self.api_client.post(url)
        return res["details"]

    def list_data_connectors(self) -> str | pd.DataFrame:
        """List the data connectors"""
        url = build_list_data_connector_url(
            LIST_DATA_CONNECTORS, None, self.organization_id
        )
        res = self.api_client.post(url)

        if res["success"]:
            df = pd.DataFrame(res["details"])
            df = df.drop(
                [
                    "_id",
                    "region",
                    "gcp_project_name",
                    "gcp_project_id",
                    "gdrive_file_name",
                    "project_name",
                ],
                axis=1,
                errors="ignore",
            )
            return df

        return res["details"]

    def list_data_connectors_buckets(self, data_connector_name) -> str | List:
        """List the buckets in data connectors

        :param data_connector_name: str
        """
        if not data_connector_name:
            return "Missing argument data_connector_name"
        if not self.organization_id:
            return "No Organization id found"

        url = build_url(LIST_BUCKETS, data_connector_name, None, self.organization_id)
        res = self.api_client.get(url)

        if res.get("message", None):
            print(res["message"])
        return res["details"]

    def list_data_connectors_filepath(
        self,
        data_connector_name,
        bucket_name: Optional[str] = None,
        root_folder: Optional[str] = None,
    ) -> str | Dict:
        """List the filepaths in data connectors

        :param data_connector_name: str
        :param bucket_name: str | Required for S3 & GCS
        :param root_folder: str | Root folder of SFTP
        """
        if not data_connector_name:
            return "Missing argument data_connector_name"
        if not self.organization_id:
            return "No Organization id found"

        def get_connector() -> str | pd.DataFrame:
            url = build_list_data_connector_url(
                LIST_DATA_CONNECTORS, None, self.organization_id
            )
            res = self.api_client.post(url)

            if res["success"]:
                df = pd.DataFrame(res["details"])
                filtered_df = df.loc[df["link_service_name"] == data_connector_name]
                if filtered_df.empty:
                    return "No data connector found"
                return filtered_df

            return res["details"]

        connectors = get_connector()
        if isinstance(connectors, pd.DataFrame):
            value = connectors.loc[
                connectors["link_service_name"] == data_connector_name,
                "link_service_type",
            ].values[0]
            ds_type = value

            if ds_type == "s3" or ds_type == "gcs":
                if not bucket_name:
                    return "Missing argument bucket_name"

            if ds_type == "sftp":
                if not root_folder:
                    return "Missing argument root_folder"

        if self.organization_id:
            url = f"{LIST_FILEPATHS}?organization_id={self.organization_id}&link_service_name={data_connector_name}&bucket_name={bucket_name}&root_folder={root_folder}"
        res = self.api_client.get(url)

        if res.get("message", None):
            print(res["message"])
        return res["details"]

    def credits(self):
        url = build_list_data_connector_url(
            COMPUTE_CREDIT_URI, None, self.organization_id
        )
        res = self.api_client.get(url)
        return res["details"]



import json
from typing import Dict, Any, Optional, Callable, List, Union
from datetime import datetime
from autogen import ConversableAgent, UserProxyAgent
from autogen_agentchat.agents import AssistantAgent
import time
from aryaxai.core.project import Project
from aryaxai.common.xai_uris import (
    RUN_GUARDRAILS_URI,
)
from opentelemetry import trace, context
from .guard_template import Guard


class GuardrailRunResult(Dict[str, Any]):
    pass  # TypedDict not needed for runtime, but can add if desired

class GuardrailSupervisor:
    """Pluggable class to monitor and control agent behavior using API-based guardrails."""
    def __init__(self, 
                 guards: Union[List[Dict[str, Any]], Dict[str, Any], None] = None,
                 apply_to: str = 'both',
                 action: str = "block",
                 project: Optional[Project] = None,
                 llm: Optional[Any] = None, 
                 ):
        if apply_to not in ['input', 'output', 'both']:
            raise ValueError("apply_to must be one of 'input', 'output', 'both'")
        self.apply_to = apply_to
        if isinstance(guards, dict):
            guards = [guards]
        self.guards = guards or []
        if action not in ['block', 'retry', 'warn']:
            raise ValueError("action must be one of 'block', 'retry', 'warn'")
        self.action = action
        if project is not None:
            self.api_client = project.api_client
            self.project_name = project.project_name
        self.llm = llm
        self.max_retries = 1
        self.retry_delay = 1.0
        self.tracer = trace.get_tracer("autogen-app")  # Standardized tracer name


    def _process_content(
        self,
        content: str,
        agent_id: str,
        content_type: str,
        action: str,
        guards: List[Dict[str, Any]],
    ) -> str:
        if not guards:
            return content

        current_content = content
        for guard in guards:
            if isinstance(guard, str):
                guard_spec: Dict[str, Any] = {"name": guard}
            else:
                guard_spec = dict(guard)

            current_content = self._apply_guardrail_with_retry(
                content=current_content,
                guard_spec=guard_spec,
                agent_id=agent_id,
                content_type=content_type,
                action=action,
            )

        return current_content

    def _apply_guardrail_with_retry(
        self,
        content: str,
        guard_spec: Dict[str, Any],
        agent_id: str,
        content_type: str,
        action: str,
    ) -> str:
        current_content = content
        retry_count = 0

        if action == "retry":
            while retry_count <= self.max_retries:
                run_result = self._call_run_guardrail(current_content, guard_spec, content_type)
                validation_passed = bool(run_result.get("validation_passed", True))
                detected_issue = not validation_passed or not run_result.get("success", True)

                if detected_issue and self.llm is not None and retry_count < self.max_retries:
                    retry_count += 1
                    time.sleep(self.retry_delay)
                    continue
                else:
                    return self._handle_action(
                        original=current_content,
                        run_result=run_result,
                        action=f"retry_{retry_count}" if retry_count > 0 else action,
                        agent_id=agent_id,
                        content_type=content_type,
                        guard_name=guard_spec.get("name", "unknown"),
                    )
        else:
            run_result = self._call_run_guardrail(current_content, guard_spec, content_type)
            return self._handle_action(
                original=current_content,
                run_result=run_result,
                action=action,
                agent_id=agent_id,
                content_type=content_type,
                guard_name=guard_spec.get("name", "unknown"),
            )

    def _call_run_guardrail(self, input_data: str, guard: Dict[str, Any], content_type: str) -> GuardrailRunResult:
        start_time = datetime.now()
        try:
            body = {"input_data": input_data, "guard": guard}
            data = self.api_client.post(RUN_GUARDRAILS_URI, body)
            # print(data , "api ran")
            end_time = datetime.now()
            details = data.get("details", {}) if isinstance(data, dict) else {}
            result: GuardrailRunResult = {
                "success": bool(data.get("success", False)) if isinstance(data, dict) else False,
                "details": details if isinstance(details, dict) else {},
                "start_time": start_time.isoformat(),
                "end_time": end_time.isoformat(),
            }
            if "duration" not in details:
                result["duration"] = (end_time - start_time).total_seconds()
            if isinstance(details, dict):
                result.update({k: v for k, v in details.items() if k in [
                    "validated_output", "validation_passed", "sanitized_output", "duration", "latency"
                ]})
            result["retry_count"] = 0
            result["max_retries"] = self.max_retries
            result["response"] = data
            result["input"] = input_data
            return result
        except Exception as exc:
            end_time = datetime.now()
            result = {
                "success": False,
                "details": {},
                "start_time": start_time.isoformat(),
                "end_time": end_time.isoformat(),
                "duration": (end_time - start_time).total_seconds(),
                "error": str(exc)
            }
            return result

    def _handle_action(
        self,
        original: str,
        run_result: GuardrailRunResult,
        action: str,
        agent_id: str,
        content_type: str,
        guard_name: str,
    ) -> str:
        validation_passed = bool(run_result.get("validation_passed", True))
        detected_issue = not validation_passed or not run_result.get("success", True)
        sanitized_output = run_result.get("sanitized_output")

        status = "passed" if not detected_issue else "failed"
        self._log_event(
            agent_id=agent_id,
            stage=content_type,
            data=original,
            status=status,
            error=run_result.get("error", ""),
            details={
                "guard": guard_name,
                "action": action,
                "detected": detected_issue,
                "duration": run_result.get("duration", 0.0),
                "input": self._safe_str(run_result.get("input")),
                "output": self._safe_str(run_result.get("response")),
            }
        )

        if detected_issue:
            on_fail_action = action
            if on_fail_action == "block":
                raise ValueError(f"Guardrail '{guard_name}' detected an issue in {content_type}. Operation blocked.")
            elif "retry" in on_fail_action:
                return original  # Return last content (would be sanitized if implemented)
            else:  # warn
                return original
        return original

    @staticmethod
    def _safe_str(value: Any) -> str:
        try:
            if isinstance(value, (str, int, float, bool)) or value is None:
                s = str(value)
                return s
            if hasattr(value, "content"):
                s = str(getattr(value, "content", ""))
                return s
            
            if isinstance(value, (list, tuple)):
                parts = []
                for item in value:
                    parts.append(Guard._safe_str(item) if hasattr(Guard, "_safe_str") else str(item))
                s = ", ".join(parts)
                return s
            
            if isinstance(value, dict):
                safe_dict: Dict[str, Any] = {}
                for k, v in value.items():
                    key = str(k)
                    if isinstance(v, (str, int, float, bool)) or v is None:
                        safe_dict[key] = v
                    elif hasattr(v, "content"):
                        safe_dict[key] = str(getattr(v, "content", ""))
                    else:
                        safe_dict[key] = str(v)
                s = json.dumps(safe_dict, ensure_ascii=False)
                return s
            s = str(value)
            return s
        except Exception:
            return "<unserializable>"

    def instrument_agents(self, agents: List[Union[ConversableAgent, AssistantAgent]]) -> List[Union[ConversableAgent, AssistantAgent]]:
        """
        Instruments a list of agents to apply guardrails.

        This method iterates through a list of agents and applies the appropriate
        instrumentation to intercept their message generation or run methods.
        It handles different agent types like `AssistantAgent` and `ConversableAgent`.

        Args:
            agents: A list of agents to be instrumented.
        
        Returns:
            The list of instrumented agents.
        """
        for agent in agents:
            # It's important to check for the more specific subclass first.
            if isinstance(agent, AssistantAgent):
                self.instrument_agent(agent)
        return agents


    async def _execute_guarded_run(self, agent, original_run, args, kwargs, current_context):
        """Execute the guarded run with proper context for guardrails."""
        # Extract task argument and ensure proper argument handling
        task = kwargs.get('task', None)
        if task is None and len(args) > 0:
            task = args[0]

        # Process input for guardrails - make them direct children of current agent span
        if self.apply_to in ['input', 'both'] and task:
            request_content = self._extract_task_content(task)
            if request_content:
                # Set input content attribute on the current agent span
                current_span = trace.get_current_span()
                if current_span.is_recording():
                    current_span.set_attribute("guardrail.input_content", self._safe_str(request_content))
                
                self._apply_input_guardrails(
                    request_content, agent.name, current_context
                )

        # Call original run method with proper argument handling
        try:
            reply = await original_run(*args, **kwargs)
                
        except Exception as e:
            raise ValueError(f"Error generating response: {str(e)}")

        # Process output through guardrails - make them direct children of current agent span
        if self.apply_to in ['output', 'both'] and reply:
            response_content = self._extract_response_content(reply)
            if response_content:
                # Set output content attribute on the current agent span
                self._apply_output_guardrails(
                    response_content, agent.name, current_context
                )

        # Ensure reply has required format for AutoGen consistency
        reply = self._format_reply(reply, agent)
        
        return reply

    def instrument_agent(self, agent) -> None:
        """Wrap AssistantAgent to intercept run method for guardrails (AutoGen 0.4+)."""
        original_run = agent.run
        # print(f"Guardrail running on {agent.__class__.__name__}")

        async def wrapped_run(*args, **kwargs):
            # print("Guardrail intercepted run method")
            
            # Get the current span and context
            current_span = trace.get_current_span()
            current_context = context.get_current()
            
            # If no parent span is active, start one for the agent run
            if not current_span.is_recording():
                with self.tracer.start_as_current_span(f"{agent.name}_run") as agent_span:
                    # Update current_context to include the new parent
                    current_context = context.get_current()
                    return await self._execute_guarded_run(agent, original_run, args, kwargs, current_context)
            else:
                return await self._execute_guarded_run(agent, original_run, args, kwargs, current_context)

        # Replace the run method with proper binding
        agent.run = wrapped_run
        return agent

    def _extract_task_content(self, task) -> str:
        """Extract content from task parameter for processing."""
        if isinstance(task, str):
            return task
        elif isinstance(task, list):
            # Handle list of messages/tasks
            content_parts = []
            for item in task:
                if isinstance(item, dict):
                    # Handle message dict format
                    if item.get('role') == 'user':
                        content_parts.append(item.get('content', ''))
                    elif 'content' in item:
                        content_parts.append(item.get('content', ''))
                    else:
                        content_parts.append(str(item))
                elif hasattr(item, 'content'):
                    content_parts.append(str(item.content))
                elif hasattr(item, 'role') and hasattr(item, 'content'):
                    if item.role == 'user':
                        content_parts.append(str(item.content))
                else:
                    content_parts.append(str(item))
            return ' '.join(filter(None, content_parts))
        elif isinstance(task, dict):
            # Handle single message dict
            return task.get('content', str(task))
        elif hasattr(task, 'content'):
            return str(task.content)
        else:
            return str(task)

    def _extract_response_content(self, reply) -> str:
        """Extract content from agent response for guardrail processing."""
        if isinstance(reply, str):
            return reply
        elif isinstance(reply, dict):
            return reply.get("content", "")
        elif hasattr(reply, "content"):
            return str(reply.content)
        elif hasattr(reply, "text"):
            return str(reply.text)
        else:
            return str(reply)

    def _apply_input_guardrails(self, content: str, agent_name: str, ctx) -> None:
        """Apply input guardrails with telemetry tracking as direct children of agent span."""
        for guard in self.guards:
            guard_name = guard.get("name", "unknown")
            
            # Create guardrail span as direct child of the current agent execution span
            with self.tracer.start_as_current_span(f"guardrail: {guard_name}", context=ctx) as guard_span:
                # Set comprehensive attributes linking this guardrail to the specific agent
                guard_span.set_attribute("component", agent_name)
                guard_span.set_attribute("guard", guard_name)
                guard_span.set_attribute("content_type", "input")
                
                try:
                    start_time = datetime.now()
                    run_result = self._call_run_guardrail(content, guard, "input")
                    end_time = datetime.now()
                    
                    # Set comprehensive telemetry attributes
                    guard_span.set_attribute("input.value", self._safe_str(content))
                    guard_span.set_attribute("output.value", 
                                        self._safe_str(run_result.get("response", "")))
                    guard_span.set_attribute("start_time", start_time.isoformat())
                    guard_span.set_attribute("end_time", end_time.isoformat())
                    guard_span.set_attribute("duration", 
                                        (end_time - start_time).total_seconds())
                    
                    # Check validation results
                    validation_passed = bool(run_result.get("validation_passed", True))
                    success = bool(run_result.get("success", True))
                    detected_issue = not validation_passed or not success
                    
                    guard_span.set_attribute("detected", detected_issue)
                    
                    # Handle guardrail violations based on your policy
                    if detected_issue:
                        guard_span.set_attribute("action", self.action)
                        error_msg = run_result.get("error_message", 
                                                f"Input guardrail '{guard_name}' detected an issue for agent '{agent_name}'")
                        guard_span.add_event("input_guardrail_violation", {
                            "agent": agent_name,
                            "guard": guard_name,
                            "error_message": error_msg
                        })
                        guard_span.set_attribute("violation.message", error_msg)
                        guard_span.set_attribute("violation.agent", agent_name)
                        # print(f"Input guardrail violation on {agent_name}: {error_msg}")
                        # Uncomment if you want to raise exceptions on violations:
                        # raise ValueError(error_msg)
                    else:
                        guard_span.set_attribute("action", "passed")
                        guard_span.add_event("input_guardrail_passed", {
                            "agent": agent_name,
                            "guard": guard_name
                        })
                        
                except Exception as e:
                    guard_span.record_exception(e)
                    guard_span.set_attribute("execution.error", True)
                    guard_span.set_attribute("error.message", str(e))
                    guard_span.set_attribute("error.agent", agent_name)
                    guard_span.add_event("input_guardrail_failed", {
                        "agent": agent_name,
                        "guard": guard_name,
                        "error": str(e)
                    })
                    # print(f"Error in input guardrail '{guard_name}' for agent '{agent_name}': {str(e)}")
                    # Re-raise if you want strict enforcement
                    # raise

    def _apply_output_guardrails(self, content: str, agent_name: str, parent_context) -> None:
        """Apply output guardrails with telemetry tracking as direct children of agent span."""
        for guard in self.guards:
            guard_name = guard.get("name", "unknown")
            
            # Create guardrail span as direct child of the current agent execution span
            with self.tracer.start_as_current_span(
                f"guardrail:{guard_name}",
                context=parent_context
            ) as guard_span:
                
                # Set comprehensive attributes linking this guardrail to the specific agent
                guard_span.set_attribute("component", agent_name)
                guard_span.set_attribute("guard", guard_name)
                guard_span.set_attribute("content_type", "output")
                
                try:
                    start_time = datetime.now()
                    guard_span.add_event("output_guardrail_started", {
                        "agent": agent_name,
                        "guard": guard_name
                    })
                    
                    run_result = self._call_run_guardrail(content, guard, "output")
                    
                    end_time = datetime.now()
                    guard_span.add_event("output_guardrail_completed", {
                        "agent": agent_name,
                        "guard": guard_name
                    })
                    
                    # Set comprehensive telemetry attributes
                    guard_span.set_attribute("input.value", self._safe_str(content))
                    guard_span.set_attribute("output.value", 
                                        self._safe_str(run_result.get("response", "")))
                    guard_span.set_attribute("start_time", start_time.isoformat())
                    guard_span.set_attribute("end_time", end_time.isoformat())
                    guard_span.set_attribute("duration", 
                                        (end_time - start_time).total_seconds())
                    
                    # Check validation results
                    validation_passed = bool(run_result.get("validation_passed", True))
                    success = bool(run_result.get("success", True))
                    detected_issue = not validation_passed or not success
                    
                    guard_span.set_attribute("detected", detected_issue)
                    
                    # Handle guardrail violations
                    if detected_issue:
                        guard_span.set_attribute("action", self.action)
                        error_msg = run_result.get("error_message", 
                                                f"Output guardrail '{guard_name}' detected an issue for agent '{agent_name}'")
                        guard_span.add_event("output_guardrail_violation", {
                            "agent": agent_name,
                            "guard": guard_name,
                            "error_message": error_msg
                        })
                        guard_span.set_attribute("violation.message", error_msg)
                        guard_span.set_attribute("violation.agent", agent_name)
                        # print(f"Output guardrail violation on {agent_name}: {error_msg}")
                        # Uncomment if you want to raise exceptions on violations:
                        # raise ValueError(error_msg)
                    else:
                        guard_span.set_attribute("action", "passed")
                        guard_span.add_event("output_guardrail_passed", {
                            "agent": agent_name,
                            "guard": guard_name
                        })
                        
                except Exception as e:
                    guard_span.record_exception(e)
                    guard_span.set_attribute("execution.error", True)
                    guard_span.set_attribute("error.message", str(e))
                    guard_span.set_attribute("error.agent", agent_name)
                    guard_span.add_event("output_guardrail_failed", {
                        "agent": agent_name,
                        "guard": guard_name,
                        "error": str(e)
                    })
                    # print(f"Error in output guardrail '{guard_name}' for agent '{agent_name}': {str(e)}")
                    # Re-raise if you want strict enforcement
                    # raise

    def _format_reply(self, reply, agent) -> Dict[str, Any]:
        """Ensure reply has consistent format for AutoGen compatibility."""
        agent_name = getattr(agent, 'name', 'assistant')
        
        if isinstance(reply, str):
            return {
                "role": "assistant", 
                "content": reply, 
                "name": agent_name
            }
        elif isinstance(reply, dict):
            # Ensure required fields exist
            formatted_reply = reply.copy()
            if "role" not in formatted_reply:
                formatted_reply["role"] = "assistant"
            if "name" not in formatted_reply:
                formatted_reply["name"] = agent_name
            if "content" not in formatted_reply and reply:
                # Try to extract content from the reply
                content = self._extract_response_content(reply)
                if content:
                    formatted_reply["content"] = content
            return formatted_reply
        else:
            # Handle other response types
            content = self._extract_response_content(reply)
            return {
                "role": "assistant",
                "content": content,
                "name": agent_name
            }

    def _safe_str(self, value, max_length: int = 1000) -> str:
        """Safely convert value to string with length limit for telemetry."""
        if value is None:
            return ""
        
        str_value = str(value)
        if len(str_value) > max_length:
            return str_value[:max_length] + "... [truncated]"
        return str_value


from opentelemetry import trace as trace_api
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor , BatchSpanProcessor
from opentelemetry.sdk.resources import Resource
from openinference.instrumentation.langchain import LangChainInstrumentor
# from openinference.instrumentation.autogen_agent import AutogenInstrumentor
# from opentelemetry.instrumentation.openai import OpenAIInstrumentor
from openinference.instrumentation.autogen_agentchat import AutogenAgentChatInstrumentor
from openinference.instrumentation.crewai import CrewAIInstrumentor
from openinference.instrumentation.openai_agents import OpenAIAgentsInstrumentor
from openinference.instrumentation.dspy import DSPyInstrumentor
from openinference.instrumentation.llama_index import LlamaIndexInstrumentor
from openinference.instrumentation.smolagents import SmolagentsInstrumentor
# from pydantic import BaseModel
import os


class Tracer:
    def __init__(self):
        self.base_url = os.getenv("XAI_API_URL", "https://apiv2.aryaxai.com")    
        self.endpoint = f"{self.base_url}"
    def setup_langchain_tracing(self , project: object, session_id : str = None) -> None:
        """
        Sets up OpenTelemetry tracing for a given project with OTLP and console exporters.

        Args:
            project: An object containing project details, expected to have a 'name' attribute.
        """
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "Langgraph",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # Instrument LangChain
        LangChainInstrumentor().instrument()
    
    def setup_autogen_tracing(self ,project: object , session_id : str = None) -> None:
        """
        Sets up OpenTelemetry tracing for a given project with OTLP and console exporters.
        
        Args:
            project: An object containing project details, expected to have a 'name' attribute.
        """
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "Autogen",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # Instrument Autogen
        AutogenAgentChatInstrumentor().instrument()

    def setup_crewai_tracing(self , project: object , session_id : str = None) -> None:
        """
        Sets up OpenTelemetry tracing for a given project with OTLP and console exporters.
        
        Args:
            project: An object containing project details, expected to have a 'name' attribute.
        """
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "Crewai",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # tracer_provider.add_span_processor(ConsoleSpanExporter())
        # Instrument CrewAI
        CrewAIInstrumentor().instrument()

    def setup_agents_tracing(self , project : object , session_id : str = None) -> None:
        """
        Sets up OpenTelemetry tracing for a given project with OTLP and console exporters.
        
        Args:
            project: An object containing project details, expected to have a 'name' attribute.
        """
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "OpenAI-Agents",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # tracer_provider.add_span_processor(ConsoleSpanExporter())
        # Instrument OpenAI
        OpenAIAgentsInstrumentor().instrument()

    def setup_dspy_tracing(self , project : object , session_id : str = None) -> None:
        """
        Sets up OpenTelemetry tracing for a given project with OTLP and console exporters.
        
        Args:
            project: An object containing project details, expected to have a 'name' attribute.
        """
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "DSPy",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # tracer_provider.add_span_processor(ConsoleSpanExporter())
        # Instrument DSPy
        DSPyInstrumentor().instrument()
    
    def setup_llamaindex_tracing(self , project : object , session_id : str = None) -> None:
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "Llamaindex",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # tracer_provider.add_span_processor(ConsoleSpanExporter())
        # Instrument llama
        LlamaIndexInstrumentor().instrument()

    def setup_smolagents_tracing(self , project : object , session_id : str = None) -> None:
        
        # Extract project name or use default
        
        project_name = getattr(project, 'project_name')
        # Create resource with service and project details
        resource = Resource(attributes={
            "service.name": "Smolagents",
            "project_name": project_name,
            "session_id": session_id if session_id else "None"
        })
        
        # Initialize tracer provider
        tracer_provider = trace_sdk.TracerProvider(resource=resource)
        trace_api.set_tracer_provider(tracer_provider)
        
        # Add OTLP and console span processors
        tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(self.endpoint)))
        # tracer_provider.add_span_processor(ConsoleSpanExporter())
        # Instrument Smol
        SmolagentsInstrumentor().instrument()


from aryaxai.common.xai_uris import (
    AVAILABLE_GUARDRAILS_URI,
    CONFIGURE_GUARDRAILS_URI,
    DELETE_GUARDRAILS_URI,
    GET_AVAILABLE_TEXT_MODELS_URI,
    GET_GUARDRAILS_URI,
    INITIALIZE_TEXT_MODEL_URI,
    MESSAGES_URI,
    SESSIONS_URI,
    TRACES_URI,
    UPDATE_GUARDRAILS_STATUS_URI,
)
from aryaxai.core.project import Project
import pandas as pd

from aryaxai.core.wrapper import  monitor


class AgentProject(Project):
    """Project for Agent modality

    :return: AgentProject
    """

    def sessions(self) -> pd.DataFrame:
        """All sessions

        :return: response
        """
        res = self.api_client.get(f"{SESSIONS_URI}?project_name={self.project_name}")
        if not res["success"]:
            raise Exception(res.get("details"))

        return pd.DataFrame(res.get("details"))

    def messages(self, session_id: str) -> pd.DataFrame:
        """All messages for a session

        :param session_id: id of the session
        :return: response
        """
        res = self.api_client.get(
            f"{MESSAGES_URI}?project_name={self.project_name}&session_id={session_id}"
        )
        if not res["success"]:
            raise Exception(res.get("details"))

        return pd.DataFrame(res.get("details"))

    def traces(self, trace_id: str) -> pd.DataFrame:
        """Traces generated for trace_id

        :param trace_id: id of the trace
        :return: response
        """
        res = self.api_client.get(
            f"{TRACES_URI}?project_name={self.project_name}&trace_id={trace_id}"
        )
        if not res["success"]:
            raise Exception(res.get("details"))

        return pd.DataFrame(res.get("details"))


from __future__ import annotations

import json
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, TypedDict, Union
from aryaxai.common.xai_uris import RUN_GUARDRAILS_URI , RUN_GUARDRAILS_PARALLEL_URI
from aryaxai.core.project import Project
from opentelemetry import trace
import time
import asyncio
import requests
from .guard_template import Guard
from dataclasses import dataclass

from agents import (
    Agent,
    RunContextWrapper,
    TResponseInputItem,
    input_guardrail,
    output_guardrail,
    ModelSettings,
    ModelTracing,
)

@dataclass
class GuardrailFunctionOutput:
    """The output of a guardrail function."""

    output_info: Any
    """
    Optional information about the guardrail's output. For example, the guardrail could include
    information about the checks it performed and granular results.
    """

    tripwire_triggered: bool
    """
    Whether the tripwire was triggered. If triggered, the agent's execution will be halted.
    """

    sanitized_content: str

class GuardrailRunResult(TypedDict, total=False):
    success: bool
    details: Dict[str, Any]
    validated_output: Any
    validation_passed: bool
    sanitized_output: Any
    duration: float
    latency: str
    on_fail_action: str
    retry_count: int
    max_retries: int
    start_time: str
    end_time: str

class OpenAIAgentsGuardrail:
    """
    Utility for creating OpenAI Agents guardrails that call Guardrails HTTP APIs.
    
    Supports different actions:
    - "block": raises InputGuardrailTripwireTriggered/OutputGuardrailTripwireTriggered
    - "retry": attempts to sanitize content using LLM and passes sanitized content to handoff agent
    - "warn": logs the issue but allows content to pass through
    """

    def __init__(
        self,
        project: Optional[Project],
        model: Optional[Any] = None,
    ) -> None:
        if project is not None:
            self.client = project.api_client
            self.project_name = project.project_name
            
        self.logs: List[Dict[str, Any]] = []
        self.max_retries = 2
        self.retry_delay = 1.0 
        self.tracer = trace.get_tracer(__name__)
        self.model = model

    def create_input_guardrail(
        self,
        guards: Union[List[str], List[Dict[str, Any]], str, Dict[str, Any]],
        action: str = "block",
        name: str = "input_guardrail"
    ) -> Callable:
        """
        Create an input guardrail function for OpenAI Agents.
        
        Args:
            guards: List of guard specifications or single guard
            action: 'block' | 'retry' | 'warn'
            name: Name for the guardrail function
        """
        if isinstance(guards, (str, dict)):
            guards = [guards]

        @input_guardrail
        async def guardrail_function(
            ctx: RunContextWrapper[None], 
            agent: Agent, 
            input: str | list[TResponseInputItem]
        ) -> GuardrailFunctionOutput:
            # Convert input to string for processing
            if isinstance(input, list):
                # Handle list of input items (messages)
                input_text = ""
                for item in input:
                    if hasattr(item, 'content'):
                        input_text += str(item.content) + " "
                    else:
                        input_text += str(item) + " "
                input_text = input_text.strip()
            else:
                input_text = str(input)

            # Process through all guards in parallel
            current_content, tripwire_triggered, output_info = await self._apply_guardrail_parallel(
                content=input_text,
                guards=guards,
                guardrail_type="input",
                action=action,
                agent_name=agent.name,
            )

            return GuardrailFunctionOutput(
                output_info=output_info,
                tripwire_triggered=tripwire_triggered,
                sanitized_content=current_content if action == "retry" else input_text
            )

        # Set function name for debugging
        guardrail_function.__name__ = name
        return guardrail_function

    def create_output_guardrail(
        self,
        guards: Union[List[str], List[Dict[str, Any]], str, Dict[str, Any]],
        action: str = "block",
        name: str = "output_guardrail"
    ) -> Callable:
        """
        Create an output guardrail function for OpenAI Agents.
        
        Args:
            guards: List of guard specifications or single guard
            action: 'block' | 'retry' | 'warn'
            name: Name for the guardrail function
        """
        if isinstance(guards, (str, dict)):
            guards = [guards]

        @output_guardrail
        async def guardrail_function(
            ctx: RunContextWrapper, 
            agent: Agent, 
            output: Any
        ) -> GuardrailFunctionOutput:
            # Extract text content from output
            if hasattr(output, 'response'):
                output_text = str(output.response)
            elif hasattr(output, 'content'):
                output_text = str(output.content)
            else:
                output_text = str(output)

            # Process through all guards in parallel
            current_content, tripwire_triggered, output_info = await self._apply_guardrail_parallel(
                content=output_text,
                guards=guards,
                guardrail_type="output",
                action=action,
                agent_name=agent.name,
            )

            return GuardrailFunctionOutput(
                output_info=output_info,
                tripwire_triggered=tripwire_triggered,
                sanitized_content=current_content if action == "retry" else output_text
            )

        # Set function name for debugging
        guardrail_function.__name__ = name
        return guardrail_function

    async def _apply_guardrail_parallel(
        self,
        content: Any,
        guards: List[Union[str, Dict[str, Any]]],
        guardrail_type: str,
        action: str,
        agent_name: str,
    ) -> tuple[Any, bool, Dict[str, Any]]:
        """
        Run multiple guardrails in parallel using batch endpoint.
        
        Returns:
            tuple: (processed_content, tripwire_triggered, output_info)
        """
        current_content = content
        tripwire_triggered = False
        output_info = {}
        retry_count = 0

        try:
            parent_span = trace.get_current_span()
            if parent_span is not None:
                ctx = trace.set_span_in_context(parent_span)
                with self.tracer.start_as_current_span(f"guardrails:{guardrail_type}", context=ctx) as parent_gr_span:
                    parent_gr_span.set_attribute("component", str(agent_name))
                    parent_gr_span.set_attribute("content_type", guardrail_type)

                    while retry_count <= self.max_retries:
                        # Prepare payload for parallel guardrail execution
                        guard_specs = [guard if isinstance(guard, dict) else {"name": guard} for guard in guards]
                        payload = {
                            "input_data": current_content,
                            "guards": guard_specs
                        }

                        # Call parallel guardrail endpoint
                        start_time = datetime.now()
                        response = self.client.post(
                            RUN_GUARDRAILS_PARALLEL_URI,
                            payload=payload,
                        )
                        end_time = datetime.now()
                        parallel_result = response

                        # Add timing information
                        parallel_result.update({
                            "start_time": start_time.isoformat(),
                            "end_time": end_time.isoformat(),
                            "duration": (end_time - start_time).total_seconds()
                        })

                        parent_gr_span.set_attribute("start_time", str(parallel_result.get("start_time", "")))
                        parent_gr_span.set_attribute("end_time", str(parallel_result.get("end_time", "")))
                        parent_gr_span.set_attribute("duration", float(parallel_result.get("duration", 0.0)))

                        if not parallel_result.get("success", False):
                            output_info["parallel_execution_error"] = parallel_result.get("details", {})
                            return current_content, tripwire_triggered, output_info

                        # Process each guardrail result
                        detected_issue = False
                        for guard_result in parallel_result.get("details", []):
                            guard_name = guard_result.get("name", "unknown")
                            run_result: GuardrailRunResult = {
                                "success": parallel_result.get("success"),
                                "details": guard_result,
                                "validated_output": guard_result.get("validated_output"),
                                "validation_passed": guard_result.get("validation_passed", False),
                                "sanitized_output": guard_result.get("sanitized_output", current_content),
                                "duration": guard_result.get("duration", 0.0),
                                "latency": guard_result.get("latency", "0 ms"),
                                "start_time": parallel_result.get("start_time", ""),
                                "end_time": parallel_result.get("end_time", ""),
                                "retry_count": retry_count,
                                "max_retries": self.max_retries
                            }
                            run_result["response"] = guard_result
                            run_result["input"] = current_content

                            # Log and handle each guard result
                            current_content, is_triggered = await self._handle_action(
                                original=current_content,
                                run_result=run_result,
                                action=f"retry_{retry_count}" if retry_count > 0 else action,
                                agent_name=agent_name,
                                guardrail_type=guardrail_type,
                                guard_name=guard_name,
                                parent_span=parent_gr_span
                            )

                            if is_triggered:
                                detected_issue = True
                                tripwire_triggered = True
                                output_info[f"guard_{guard_name}"] = run_result

                        if detected_issue and action == "retry" and self.model is not None and retry_count < self.max_retries:
                            # Sanitize content using LLM
                            prompt = self._build_sanitize_prompt("combined", current_content, guardrail_type)
                            try:
                                sanitized = await self._invoke_llm(prompt)
                                current_content = sanitized
                            except Exception:
                                pass  # Keep original content if sanitization fails
                            retry_count += 1
                            await self._async_sleep(self.retry_delay)
                            continue
                        else:
                            # No issues or no retries left
                            output_info["retry_count"] = retry_count
                            output_info["final_content"] = current_content
                            return current_content, tripwire_triggered, output_info

            # Fallback if no parent span
            output_info["retry_count"] = retry_count
            output_info["final_content"] = current_content
            return current_content, tripwire_triggered, output_info

        except Exception as e:
            output_info["error"] = f"Parallel guardrail execution failed: {str(e)}"
            return current_content, tripwire_triggered, output_info

    async def _handle_action(
        self,
        original: Any,
        run_result: GuardrailRunResult,
        action: str,
        agent_name: str,
        guardrail_type: str,
        guard_name: str,
        parent_span: Optional[Any]
    ) -> tuple[Any, bool]:
        """
        Handle the action for a single guardrail result.
        
        Returns:
            tuple: (processed_content, is_triggered)
        """
        validation_passed = bool(run_result.get("validation_passed", True))
        detected_issue = not validation_passed or not run_result.get("success", True)

        if parent_span is not None:
            try:
                with self.tracer.start_as_current_span(
                    f"guard:{guard_name}",
                    context=trace.set_span_in_context(parent_span)
                ) as gr_span:
                    gr_span.set_attribute("component", str(agent_name))
                    gr_span.set_attribute("guard", str(guard_name))
                    gr_span.set_attribute("content_type", guardrail_type)
                    gr_span.set_attribute("detected", detected_issue)
                    gr_span.set_attribute("action", action)
                    gr_span.set_attribute("input.value", self._safe_str(run_result.get("input")))
                    gr_span.set_attribute("output.value", json.dumps(run_result.get("response")))
                    gr_span.set_attribute("start_time", str(run_result.get("start_time", "")))
                    gr_span.set_attribute("end_time", str(run_result.get("end_time", "")))
                    gr_span.set_attribute("duration", float(run_result.get("duration", 0.0)))
            except Exception:
                pass

        # Log guardrail result without creating a new span
        self._log_guardrail_result(
            run_result=run_result,
            action=action,
            agent_name=agent_name,
            guardrail_type=guardrail_type,
            guard_name=guard_name,
        )

        if detected_issue:
            if action == "block":
                return original, True
            elif "retry" in action:
                return run_result.get("sanitized_output", original), True
            else:  # warn
                return original, False
        return original, False

    async def _call_run_guardrail(
        self, 
        input_data: Any, 
        guard: Dict[str, Any], 
        guardrail_type: str
    ) -> GuardrailRunResult:
        """Call the guardrails HTTP API"""
        uri = RUN_GUARDRAILS_URI
        input_text = str(input_data)
        
        start_time = datetime.now()
        try:
            body = {"input_data": input_text, "guard": guard}
            data = self.client.post(uri, body)
            
            end_time = datetime.now()
            
            details = data.get("details", {}) if isinstance(data, dict) else {}
            result: GuardrailRunResult = {
                "success": bool(data.get("success", False)) if isinstance(data, dict) else False,
                "details": details if isinstance(details, dict) else {},
                "start_time": start_time.isoformat(),
                "end_time": end_time.isoformat(),
            }
            
            if "duration" not in details:
                result["duration"] = (end_time - start_time).total_seconds()
                
            if isinstance(details, dict):
                if "validated_output" in details:
                    result["validated_output"] = details["validated_output"]
                if "validation_passed" in details:
                    result["validation_passed"] = details["validation_passed"]
                if "sanitized_output" in details:
                    result["sanitized_output"] = details["sanitized_output"]
                if "duration" in details:
                    result["duration"] = details["duration"]
                if "latency" in details:
                    result["latency"] = details["latency"]
                
            result["retry_count"] = 0
            result["max_retries"] = self.max_retries
            result["response"] = data
            result["input"] = input_text

            return result
            
        except Exception as exc:
            end_time = datetime.now()
            raise exc

    def _log_guardrail_result(
        self,
        run_result: GuardrailRunResult,
        action: str,
        agent_name: str,
        guardrail_type: str,
        guard_name: str,
    ) -> None:
        """Log guardrail results without creating a new span"""
        validation_passed = bool(run_result.get("validation_passed", True))
        detected_issue = not validation_passed or not run_result.get("success", True)
        
        # Append to logs list instead of creating a new span
        log_entry = {
            "guard_name": guard_name,
            "guardrail_type": guardrail_type,
            "agent_name": agent_name,
            "action": action,
            "detected_issue": detected_issue,
            "start_time": run_result.get("start_time", ""),
            "end_time": run_result.get("end_time", ""),
            "duration": float(run_result.get("duration", 0.0)),
            "input": self._safe_str(run_result.get("input")),
            "output": self._safe_str(run_result.get("response"))
        }
        self.logs.append(log_entry)

    def _build_sanitize_prompt(self, guard_name: str, content: Any, guardrail_type: str) -> str:
        """Build a prompt for the LLM to sanitize the content according to the guardrail type"""
        instructions = {
            "Detect PII": "Sanitize the following text by removing or masking any personally identifiable information (PII). Do not change anything else.",
            "NSFW Text": "Sanitize the following text by removing or masking any not safe for work (NSFW) content. Do not change anything else.",
            "Ban List": "Sanitize the following text by removing or masking any banned words. Do not change anything else.",
            "Bias Check": "Sanitize the following text by removing or masking any biased language. Do not change anything else.",
            "Competitor Check": "Sanitize the following text by removing or masking any competitor names. Do not change anything else.",
            "Correct Language": "Sanitize the following text by correcting the language to the expected language. Do not change anything else.",
            "Gibberish Text": "Sanitize the following text by removing or correcting any gibberish. Do not change anything else.",
            "Profanity Free": "Sanitize the following text by removing or masking any profanity. Do not change anything else.",
            "Secrets Present": "Sanitize the following text by removing or masking any secrets. Do not change anything else.",
            "Toxic Language": "Sanitize the following text by removing or masking any toxic language. Do not change anything else.",
            "Contains String": "Sanitize the following text by removing or masking the specified substring. Do not change anything else.",
            "Detect Jailbreak": "Sanitize the following text by removing or masking any jailbreak attempts. Do not change anything else.",
            "Endpoint Is Reachable": "Sanitize the following text by ensuring any mentioned endpoints are reachable. Do not change anything else.",
            "Ends With": "Sanitize the following text by ensuring it ends with the specified string. Do not change anything else.",
            "Has Url": "Sanitize the following text by removing or masking any URLs. Do not change anything else.",
            "Lower Case": "Sanitize the following text by converting it to lower case. Do not change anything else.",
            "Mentions Drugs": "Sanitize the following text by removing or masking any mentions of drugs. Do not change anything else.",
            "One Line": "Sanitize the following text by ensuring it is a single line. Do not change anything else.",
            "Reading Time": "Sanitize the following text by ensuring its reading time matches the specified value. Do not change anything else.",
            "Redundant Sentences": "Sanitize the following text by removing redundant sentences. Do not change anything else.",
            "Regex Match": "Sanitize the following text by ensuring it matches the specified regex. Do not change anything else.",
            "Sql Column Presence": "Sanitize the following text by ensuring specified SQL columns are present. Do not change anything else.",
            "Two Words": "Sanitize the following text by ensuring it contains only two words. Do not change anything else.",
            "Upper Case": "Sanitize the following text by converting it to upper case. Do not change anything else.",
            "Valid Choices": "Sanitize the following text by ensuring it matches one of the valid choices. Do not change anything else.",
            "Valid Json": "Sanitize the following text by ensuring it is valid JSON. Do not change anything else.",
            "Valid Length": "Sanitize the following text by ensuring its length is valid. Do not change anything else.",
            "Valid Range": "Sanitize the following text by ensuring its value is within the valid range. Do not change anything else.",
            "Valid URL": "Sanitize the following text by ensuring it is a valid URL. Do not change anything else.",
            "Web Sanitization": "Sanitize the following text by removing any unsafe web content. Do not change anything else.",
        }
        
        instruction = instructions.get(guard_name, "Sanitize the following text according to the guardrail requirements. Do not change anything else.")
        prompt = f"{instruction}\n\nContent:\n{content}"
        return prompt

    async def _invoke_llm(self, prompt: str) -> str:
        """Invoke the LLM for content sanitization"""
        if self.model is None:
            return prompt  # Return original if no LLM available
            
        try:
            data = await self.model.get_response(
                system_instructions="Based on the input you have to provide the best and accurate results",
                input=prompt,
                model_settings=ModelSettings(temperature=0.1),
                tools=[],
                output_schema=None,
                handoffs=[],
                tracing=ModelTracing.DISABLED,
                previous_response_id=None
            )
            return str(data.output[0].content[0].text)
        except Exception:
            return prompt  # Return original on error

    async def _async_sleep(self, seconds: float) -> None:
        """Async sleep utility"""
        await asyncio.sleep(seconds)

    @staticmethod
    def _safe_str(value: Any) -> str:
        """Safely convert any value to string for logging"""
        try:
            if isinstance(value, (str, int, float, bool)) or value is None:
                return str(value)
            if hasattr(value, "content"):
                return str(getattr(value, "content", ""))
            
            if isinstance(value, (list, tuple)):
                parts = []
                for item in value:
                    parts.append(OpenAIAgentsGuardrail._safe_str(item))
                return ", ".join(parts)
            
            if isinstance(value, dict):
                safe_dict: Dict[str, Any] = {}
                for k, v in value.items():
                    key = str(k)
                    if isinstance(v, (str, int, float, bool)) or v is None:
                        safe_dict[key] = v
                    elif hasattr(v, "content"):
                        safe_dict[key] = str(getattr(v, "content", ""))
                    else:
                        safe_dict[key] = str(v)
                return json.dumps(safe_dict, ensure_ascii=False)
            
            return str(value)
        except Exception:
            return "<unserializable>"

def create_guardrail(project: Project, model: Optional[Any] = None) -> OpenAIAgentsGuardrail:
    """Quick factory function to create a guardrail instance with a project"""
    return OpenAIAgentsGuardrail(project=project, model=model)